# mypy: disable-error-code="method-assign"
from __future__ import annotations

import io
import platform
from contextlib import contextmanager
from pathlib import Path

try:
    import plotly.io as plotly_io

    PLOTLY_AVAIL = True
except (ImportError, ModuleNotFoundError):
    PLOTLY_AVAIL = False

from pharaoh import project
from pharaoh.assetlib.context import context_stack
from pharaoh.assetlib.util import parse_signature

if PLOTLY_AVAIL:
    try:
        import kaleido  # noqa: F401
    except (ImportError, ModuleNotFoundError):
        if platform.platform() == "Windows":
            msg = (
                "The 'kaleido' package is missing. It is required to export static PNGs from plotly plots. "
                "Please install it into the current environment."
            )
            raise Exception(msg) from None

    # memorize the unpatched functions
    vanilla_px_show = plotly_io.show
    vanilla_px_write_image = plotly_io.write_image
    vanilla_px_write_html = plotly_io.write_html

    @contextmanager
    def patch():
        """
        A context manager that patches all functions of plotly relevant to display/save figures when executed in a
        pharaoh context. When the program exits the patch-context, the patching will be undone.

        The main goals:
            - retain original function call signature
            - suppress display of figures in GUI or browser
            - alter storage location of figures to the pharaoh build folder
            - deduplicate shared plot resources (js/css files)
            - convert all html plots into static image generation by the flick of a switch
            - if static image output is forced, inject static image parameters via config file
        """

        # switching out the original with the patched functions
        if plotly_io.show == vanilla_px_show:
            plotly_io.show = patched_px_show
            plotly_io.write_image = patched_px_write_image
            plotly_io.write_html = patched_px_write_html

        yield

        # undo the patching
        if plotly_io.show != vanilla_px_show:
            plotly_io.show = vanilla_px_show
            plotly_io.write_image = vanilla_px_write_image
            plotly_io.write_html = vanilla_px_write_html

    def patched_px_show(*args, **kwargs):
        """
        This is the patched version of ``plotly.graph_objects.Figure.write_image(*args, **kwargs)``.

        Changes:
            - becomes a no-op when executed in a pharaoh context

        docs: https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html#plotly.graph_objects.Figure.show
        """

    def patched_px_write_image(*args, **kwargs):
        """
        This is the patched version of ``plotly.graph_objects.Figure.write_image(...)`` function.

        Changes:
            - takes kwargs from the pharaoh config file and updates the kwargs passed into the function
            - changes the plots storage destination to be inside pharaoh's asset build folder

        docs: https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html#plotly.graph_objects.Figure.write_image
        """
        params = parse_signature(obj="plotly.io._kaleido.write_image", args=args, kwargs=kwargs)
        file = params["file"]
        params.pop("format")  # Format must be given through filename
        params.pop("engine")  # Enforce auto

        if isinstance(file, io.IOBase) or not Path(file).suffix:
            msg = "The file argument must be a valid file path (including suffix)!"
            raise Exception(msg)

        active_app = project.get_project()

        format = Path(file).suffix.lower()
        if format in (
            ".jpeg",
            ".jpg",
            ".png",
            ".svg",
        ):
            template_context = {"template": "image"}
        else:
            msg = f"File format of {file} not supported!"
            raise NotImplementedError(msg)

        settings = active_app.get_setting("toolkits.plotly.write_image")
        for tag in ("width", "height", "scale", "validate"):
            if tag not in params or params[tag] is None:
                params[tag] = getattr(params["fig"].layout, tag, None) or settings.get(tag)

        file_path = active_app._build_asset_filepath(file)
        params["file"] = str(file_path)
        with context_stack.new_context(
            context_name="plotly",
            asset=dict(
                user_filepath=str(file),
                file=str(file_path),
                name=file_path.name,
                stem=file_path.stem,
                suffix=file_path.suffix,
                **template_context,
            ),
        ):
            context_stack.get_parent_context(name="generate_assets")["asset"]["index"] += 1
            vanilla_px_write_image(**params)
            context_stack.dump(file_path)

    def patched_px_write_html(*args, **kwargs):
        """
        This is the patched version of ``plotly.graph_objects.Figure.write_html(*args, **kwargs)`` function.

        Changes:
            - takes kwargs from the pharaoh config file and updates the kwargs passed into the function
            - changes the plots storage destination to be inside pharaoh's asset build folder

        docs: https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html#plotly.graph_objects.Figure.write_html
        """

        params = parse_signature(obj="plotly.io._html.write_html", args=args, kwargs=kwargs)
        params["include_plotlyjs"] = "cdn"
        params["include_mathjax"] = "cdn"
        params["full_html"] = False
        file = params["file"]

        if isinstance(file, io.IOBase) or not Path(file).suffix:
            msg = "The file argument must be a valid file path (including suffix)!"
            raise Exception(msg)

        active_app = project.get_project()

        if active_app.get_setting("asset_gen.force_static"):
            coerced_path = Path(file).with_suffix(".png")
            _kwargs = {"fig": params["fig"], "file": str(coerced_path)}
            patched_px_write_image(**_kwargs)
            return

        file_path = active_app._build_asset_filepath(file)
        settings = active_app.get_setting("toolkits.plotly.write_html")
        for tag in ("default_width", "default_height"):
            params.setdefault(tag, settings.get(tag))
        params["file"] = str(file_path)

        # Set a default size
        with context_stack.new_context(
            context_name="plotly",
            asset={
                "user_filepath": str(file),
                "file": str(file_path),
                "name": file_path.name,
                "stem": file_path.stem,
                "suffix": file_path.suffix,
                "template": "raw_html",
            },
        ):
            context_stack.get_parent_context(name="generate_assets")["asset"]["index"] += 1
            vanilla_px_write_html(**params)
            context_stack.dump(file_path)

else:
    # if the package is not installed there is nothing to patch, so we silently do nothing when patch() is called.

    @contextmanager
    def patch():
        yield



from __future__ import annotations

from typing import Callable

env_tests: dict[str, Callable] = {}


# Only document the functions defined in here
to_document = {k: func for k, func in env_tests.items() if hasattr(func, "__module__") and func.__module__ == __name__}



from __future__ import annotations

import contextlib
import io
import traceback

import pharaoh.log
from pharaoh.assetlib.generation import register_asset

log = pharaoh.log.log


@contextlib.contextmanager
def catch_exceptions(
    catch: tuple[type[Exception], ...] = (Exception,),
    reraise: tuple[type[Exception], ...] = (),
    msg_prefix: str = "",
    log_exc: bool = True,
    render_exc: bool = True,
):
    """
    Catch exceptions of given types.

    If an exception is caught, the Sphinx report generation will report a warning.

    .. seealso:: :ref:`reference/assets:Catching Errors`

    :param catch: The exception types to catch
    :param reraise: The exception types to re-raise (has precedence over catch)
    :param msg_prefix: The message prefix to log before the exception message
    :param log_exc: Whether to log the exception message as a warning
    :param render_exc: Whether to export the exception traceback as an asset to be included in the report
    """
    try:
        yield
    except reraise:
        raise
    except catch as e:
        if log_exc:
            log.warning(
                f"An error occurred during asset generation: {msg_prefix}{e}\n"
                f"Traceback:\n{traceback.format_exc(limit=-1)}"
            )
        if render_exc:
            register_asset(
                file="error.txt",
                template="error_traceback",
                data=io.BytesIO(traceback.format_exc(limit=-1).encode("utf-8")),
                metadata={"error_message": f"{msg_prefix}{e}", "asset_type": "error_traceback"},
            )



from __future__ import annotations

import ast
import json
import tempfile
import typing as t
from pathlib import Path

from .find import find_template
from .scaffolder import Scaffolder


def render_template(template_path: Path, outputdir: Path, context: dict | None = None) -> Path:
    if template_path.is_dir():
        return render_template_directory(template_path, outputdir, context)
    if template_path.is_file():
        return render_template_file(template_path, outputdir, context)
    raise NotImplementedError


def render_template_directory(template_path: Path, outputdir: Path, context: dict | None = None) -> Path:
    outputdir.mkdir(parents=True, exist_ok=True)
    assert template_path.is_dir(), f"{template_path} is not a directory!"

    Scaffolder(source_dir=template_path, target_dir=outputdir, render_context=context).run()

    if isinstance(context, dict):
        context_dump_file = outputdir / ".template_context.json"
        context_dump_file.write_text(json.dumps(context, indent=2))
    return outputdir


def render_template_file(template_path: Path, outputdir: Path, context: dict | None = None) -> Path:
    if [suffix.lower() for suffix in template_path.suffixes] != [".pharaoh", ".py"]:
        msg = "If template_path is a file, only files with .pharaoh.py suffix are supported!"
        raise ValueError(msg)

    with tempfile.TemporaryDirectory(prefix="pharaoh_") as tdir:
        tempdir = Path(tdir)
        template_content = template_path.read_text(encoding="utf-8")
        tree = ast.parse(template_content)
        module_docstring = (ast.get_docstring(tree) or "").strip()

        stem = ".".join(template_path.name.split(".")[:-2])
        asset_scripts = tempdir / "asset_scripts"
        asset_scripts.mkdir()
        (asset_scripts / f"{stem}.py").write_text(template_content, encoding="utf-8")
        if module_docstring:
            (tempdir / f"index_{stem}.rst").write_text(module_docstring, encoding="utf-8")

        return render_template_directory(tempdir, outputdir, context)


def render_sphinx_base_project(outputdir: Path, templates: t.Iterable[str], context: dict | None = None):
    for templ in templates:
        render_template_directory(
            template_path=find_template(templ),
            outputdir=outputdir,
            context=context,
        )



# mypy: disable-error-code="method-assign"
from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path

try:
    import panel.io.save as panel_io_save
    from bokeh.core.templates import get_env

    PANEL_AVAIL = True
except (ImportError, ModuleNotFoundError):
    PANEL_AVAIL = False

# from pharaoh import project
from pharaoh.assetlib.util import parse_signature

if PANEL_AVAIL:
    # memorize the unpatched functions
    vanilla_panel_io_save_save = panel_io_save.save

    @contextmanager
    def patch():
        # switching out the original with the patched functions
        # Not implemented yet
        # setattr(panel_io_save, "save", patched_panel_io_save_save)

        yield

        # undo the patching
        # setattr(panel_io_save, "save", vanilla_panel_io_save_save)

    def patched_panel_io_save_save(*args, **kwargs):
        """
        This is the patched version of ``panel.io.save.save(*args, **kwargs)`` function.
        """
        params = parse_signature(obj=vanilla_panel_io_save_save, args=args, kwargs=kwargs)
        params.pop("kwargs")
        params.update(kwargs)
        params["template"] = get_env().from_string(
            Path(__file__).with_name("bokeh_embed.html").read_text(encoding="utf-8")
        )
        raise NotImplementedError
        vanilla_panel_io_save_save(**params)

else:

    def init_module():
        pass

    @contextmanager
    def patch():
        yield



# mypy: disable-error-code="method-assign"
from __future__ import annotations

import io
from contextlib import contextmanager
from pathlib import Path

try:
    import holoviews as hv

    HOLOVIEWS_AVAIL = True
except (ImportError, ModuleNotFoundError):
    HOLOVIEWS_AVAIL = False

if HOLOVIEWS_AVAIL:
    # memorize the unpatched functions
    vanilla_hv_save = hv.save

    @contextmanager
    def patch():
        """
        A context manager that patches all functions of holoviews relevant to display/save figures when executed in a
        pharaoh context. When the program exits the patch-context, the patching will be undone.

        The main goals:
            - retain original function call signature
            - alter storage location of figures to the pharaoh build folder
            - convert all html plots into static image generation by the flick of a switch
            - if static image output is forced, inject static image parameters via config file
        """

        # switching out the original with the patched functions
        if hv.save == vanilla_hv_save:
            hv.save = patched_hv_save

        yield

        # undo the patching
        if hv.save != vanilla_hv_save:
            hv.save = vanilla_hv_save

    def patched_hv_save(obj, filename, fmt="auto", backend=None, resources="cdn", toolbar=None, title=None, **kwargs):
        """
        This is the patched version of ``holoviews.util.save(obj, filename, fmt='auto', backend=None, resources='cdn',
        toolbar=None, title=None, **kwargs)`` function.

        Changes:
            - if format is html but force-static-flag is set, takes kwargs from the pharaoh config file and updates
              the kwargs passed into the function
            - changes the plots storage destination to be inside pharaoh's asset build folder

        docs: https://holoviews.org/reference_manual/holoviews.util.html#holoviews.util.save
        """
        backend = (backend or hv.Store.current_backend or "plotly").lower()

        if isinstance(filename, io.IOBase) or not Path(filename).suffix:
            msg = "The file argument must be a valid file path (including suffix)!"
            raise Exception(msg)

        suffix = Path(filename).suffix.lower()

        if backend == "bokeh":
            from bokeh.io import save
            from bokeh.io.export import export_png, export_svg

            bokeh_fig = hv.render(obj, backend="bokeh")

            if suffix.lower() == ".html":
                save(bokeh_fig, filename, **kwargs)
            elif suffix.lower() in (".png", ".jpg", ".jpeg"):
                export_png(bokeh_fig, filename=filename, **kwargs)
            elif suffix.lower() == ".svg":
                export_svg(bokeh_fig, filename=filename, **kwargs)
            else:
                raise NotImplementedError

        elif backend == "plotly":
            from plotly.graph_objs import Figure

            pfig = hv.render(obj, backend="plotly")
            pfig.pop("config")
            plotly_fig = Figure(**pfig)
            if suffix.lower() == ".html":
                plotly_fig.write_html(filename, **kwargs)
            else:
                plotly_fig.write_image(filename, **kwargs)
        elif backend == "matplotlib":
            from plotly.graph_objs import Figure

            fig = hv.render(obj, backend="matplotlib")
            fig.savefig(filename, **kwargs)
        else:
            msg = f"Unsupported backend {backend!r}!"
            raise ValueError(msg)

else:
    # if the package is not installed there is nothing to patch, so we silently do nothing when patch() is called.

    @contextmanager
    def patch():
        yield



from __future__ import annotations

import datetime
import getpass

import omegaconf


def pharaoh_project_dir(*args):
    from pharaoh.api import get_project

    return get_project().project_root.as_posix()


omegaconf.OmegaConf.register_new_resolver(
    name="pharaoh.project_dir",
    resolver=pharaoh_project_dir,
)


omegaconf.OmegaConf.register_new_resolver(
    name="now.strf",
    resolver=lambda fmt: datetime.datetime.now(
        tz=datetime.datetime.now(tz=datetime.timezone.utc).astimezone().tzinfo
    ).strftime(fmt),
)


omegaconf.OmegaConf.register_new_resolver(
    name="user",
    resolver=getpass.getuser,
)

omegaconf.OmegaConf.register_new_resolver(
    name="utcnow.strf",
    resolver=lambda fmt: datetime.datetime.now(tz=datetime.timezone.utc).strftime(fmt),
)



from __future__ import annotations

import copy
from contextlib import AbstractContextManager
from functools import partial, reduce
from pathlib import Path
from typing import Union

from pharaoh.log import log
from pharaoh.util import json_encoder, mergedeep

PathLike = Union[str, Path]


class MetadataContext(AbstractContextManager):
    def __init__(self, stack: MetadataContextStack, context_name: str = "", **context):
        self._stack = stack
        self._context = context
        self._context_name = context_name

    def __enter__(self):
        self._stack.append_context(self)
        return self

    def __exit__(self, *exc_args):
        self._stack.remove_context(self)

    def activate(self):
        return self.__enter__()

    def deactivate(self):
        self.__exit__()

    def __setitem__(self, key, value):
        self._context[key] = value

    def __getitem__(self, item):
        return self._context[item]

    def __repr__(self):
        return f"{self.__class__.__name__}[{self._context_name}]: {self._context}"

    def __str__(self):
        return repr(self)

    @property
    def name(self):
        return self._context_name


class MetadataContextStack:
    def __init__(self):
        self._stack: list[MetadataContext] = []

    def append_context(self, context: MetadataContext):
        self._stack.append(context)

    def remove_context(self, context: MetadataContext):
        if context in self._stack:
            self._stack.remove(context)

    def new_context(self, context_name="", **context) -> MetadataContext:
        """
        Returns a context manager that adds a new stack entry with custom metadata,
        effectively updating the metadata context assets are exported with.

        When the context manager is left, the stack entry is removed again.

        See :ref:`Metadata Stack docs <reference/assets:Metadata Stack>`.

        :param context_name: An optional name for the context in case it must be looked up.
        :param context: Keywords arguments to specify custom metadata.
        """
        context.pop("stack", None)
        return MetadataContext(context_name=context_name, stack=self, **context)

    def get_parent_context(self, name: str = "", index: int = -1) -> MetadataContext:
        if not name:
            return self._stack[index]

        for ctx in self._stack[::-1]:
            if ctx.name == name:
                return ctx
        msg = f"Cannot find context with name {name!r}!"
        raise LookupError(msg)

    def current_context(self):
        return self._stack[-1]

    def reset(self):
        self._stack.clear()

    def merge_stacks(self) -> dict:
        if len(self._stack) > 1:
            merge = partial(mergedeep.merge, safe=False)
            return reduce(merge, [copy.deepcopy(s._context) for s in self._stack])

        if len(self._stack) == 1:
            return copy.deepcopy(self._stack[0]._context)

        return {}

    def dump(self, asset_filepath: PathLike) -> Path:
        """
        Dumps the currently active context stack to a companion file of the
        asset with the file suffix ".assetinfo" and returns its path.

        :param asset_filepath: The path to the asset file for which the companion file shall be created.
                               This path is created by PharaohApp.build_asset_filepath
        """
        asset_filepath = Path(asset_filepath)
        merged_stack = self.merge_stacks()
        assetinfo = asset_filepath.parent / f"{asset_filepath.stem}.assetinfo"
        assetinfo.write_text(json_encoder.encode_json(merged_stack, indent=1))
        log.debug(
            f"Created asset {merged_stack['asset']['name']!r} from script "
            f"{merged_stack['asset'].get('script_name', '__unknown__')!r}."
        )
        return assetinfo


context_stack = MetadataContextStack()
metadata_context = context_stack.new_context


if __name__ == "__main__":  # pragma: no cover
    stack = MetadataContextStack()
    metadata_context = stack.new_context

    with metadata_context(a=1), metadata_context(b=2):
        pass



from __future__ import annotations

import functools
import os
import pprint
import shutil
import uuid
from functools import partial
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Callable

import jinja2
import jinja2.utils
import omegaconf
from jinja2_git import GitExtension

from pharaoh.log import log
from pharaoh.util.contextlib_chdir import chdir

from .env_filters import env_filters
from .env_globals import env_globals
from .env_tests import env_tests
from .util import asset_rel_path_from_build, asset_rel_path_from_project

if TYPE_CHECKING:
    from collections.abc import Iterator

    from sphinx.config import Config

    import pharaoh.project
    from pharaoh.sphinx_app import PharaohSphinx


class PharaohFileSystemLoader(jinja2.loaders.FileSystemLoader):
    def get_source(self, environment: jinja2.Environment, template: str) -> tuple[str, str, Callable[[], bool]]:
        # Overwrite to support absolute filenames as well as relative ones that have to be looked up in the search paths
        for searchpath in self.searchpath:
            if "<>" in template:  # See PharaohTemplateEnv.join_path
                parent, template_ = template.rsplit("<>", 1)
                template_path = Path(parent) / template_
                if template_path.is_absolute() and template_path.exists():
                    filename = template_path.as_posix()
                else:
                    pieces = jinja2.loaders.split_template_path(template_)
                    filename = jinja2.loaders.posixpath.join(searchpath, *pieces)
            else:
                pieces = jinja2.loaders.split_template_path(template)
                filename = jinja2.loaders.posixpath.join(searchpath, *pieces)

            # Original code starts from here

            if not os.path.isfile(filename):
                continue
            with open(filename, "rb") as f:
                contents = f.read().decode(self.encoding)

            def up_to_date() -> bool:
                return False

            # Use normpath to convert Windows altsep to sep.
            return contents, os.path.normpath(filename), up_to_date
        raise jinja2.TemplateNotFound(template)


class PharaohTemplate(jinja2.Template):
    def render(self, *args, **kwargs) -> str:
        return super().render(*args, **kwargs)


class PharaohTemplateEnv(jinja2.Environment):
    template_class = PharaohTemplate

    def __init__(self):
        super().__init__(
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
            extensions=["jinja2_ansible_filters.AnsibleCoreFiltersExtension"],
        )
        self.default_context: dict = {
            "project": {},  # Project related context
            "local": {},  # Discovered content of context files next to the source file
            "assets": {},  # Discovered content of asset files registered via register_templating_context function
            "config": None,  # Content of conf.py (Sphinx Config object)
            "user": None,  # Content of user given dict "pharaoh_jinja_context" in conf.py
        }
        self.local_context_file_cache: dict[Path, ModuleType] = {}

        self.sphinx_app: PharaohSphinx | None = None
        self.globals.update(env_globals)
        self.filters.update(env_filters)
        self.tests.update(env_tests)

        # todo: remove or enable if needed. Unused right now
        # from jinjatag.extension import JinjaTag
        # jinja_tag = JinjaTag()
        # self.add_extension(jinja_tag)
        # jinja_tag.init()

        self.add_extension(GitExtension)

    def sphinx_config_inited_hook(self, app: PharaohSphinx, config: Config):
        """
        Called by Sphinx core event "config-inited". Emitted when the config object has been initialized.
        https://www.sphinx-doc.org/en/master/extdev/appapi.html#event-config-inited
        """
        from pharaoh.plugins.plugin_manager import PM

        self.sphinx_app = app
        pharaoh_proj = app.pharaoh_proj

        # Configure globals/filters from Pharaoh app
        pharaoh_proj._update_template_env(self)

        # Configure users globals/filters/tests from conf.py
        user_filters = self.sphinx_app.config.pharaoh_jinja_filters or {}
        for k, v in user_filters.items():
            if k in self.filters:
                msg = f"A Jinja filter named {k!r} already exists: {self.filters[k]}"
                raise NameError(msg)
            self.filters[k] = v

        user_globals = self.sphinx_app.config.pharaoh_jinja_globals or {}
        for k, v in user_globals.items():
            if k in self.globals:
                msg = f"A Jinja global named {k!r} already exists: {self.globals[k]}"
                raise NameError(msg)
            self.globals[k] = v

        user_tests = self.sphinx_app.config.pharaoh_jinja_tests or {}
        for k, v in user_tests.items():
            if k in self.tests:
                msg = f"A Jinja test named {k!r} already exists: {self.tests[k]}"
                raise NameError(msg)
            self.tests[k] = v

        template_paths = PM.pharaoh_collect_l2_templates()
        # dynamically discovered
        for path in config.pharaoh_jinja_templates or []:
            # Extend user-configured template paths
            path = Path(path)
            if not path.is_absolute():
                if str(path) not in self.sphinx_app.config.exclude_patterns:
                    # Extend exclude patterns by user's local template directory, so Sphinx does not warn about files
                    # that are not included in any toctree.
                    self.sphinx_app.config.exclude_patterns.append(str(path))
                p = Path(self.sphinx_app.confdir) / path
                if not p.exists() and not p.is_dir():
                    msg = f"Directory {p} does not exists or is no directory!"
                    raise FileNotFoundError(msg)
                template_paths.append(p)
            else:
                template_paths.append(path.resolve())

        self.loader = PharaohFileSystemLoader(template_paths)

        self.default_context["project"]["instance"] = pharaoh_proj
        self.default_context["config"] = app.config or {}
        self.default_context["user"] = app.config.pharaoh_jinja_context or {}

    def sphinx_builder_inited_hook(self, app: PharaohSphinx):
        """
        Called by Sphinx core event "builder-inited". Emitted when the builder object has been created.
        https://www.sphinx-doc.org/en/master/extdev/appapi.html#event-builder-inited
        """
        finder = app.pharaoh_proj.asset_finder

        # Copy shared resources
        assets_to_copy = finder.search_assets("asset.copy2build")
        build_assetdir = app.assets_dir
        if build_assetdir.exists():
            shutil.rmtree(build_assetdir, ignore_errors=False)

        for asset in assets_to_copy:
            log.debug(f"Copying asset {asset} to build directory")
            asset.copy_to(build_assetdir)

    def sphinx_source_read_hook(self, app: PharaohSphinx, docname: str, source: list):
        """
        Emitted when a source file has been read. The source argument is a list whose single element is the contents
        of the source file.
        You can process the contents and replace this item to implement source-level transformations.

        :param app: The Sphinx app
        :param docname: The document name as string
        :param source: A list with a single item, the string content of 'docname'.
                       May be modified in place to change the source Sphinx will use.
                       Here it is used send the file content through Jinja rendering before Sphinx gets it.
        """
        rendered, _ = self.render_file(docname)
        source[0] = rendered  # source item 0 must be modified in place

    def sphinx_include_read_hook(self, app: PharaohSphinx, relative_path, parent_docname, content):
        file = Path(self.sphinx_app.srcdir) / relative_path
        rendered, _ = self.render_file(file)
        content[0] = rendered

    def render_file(self, docname: Path | str) -> tuple[str, Path]:
        project: pharaoh.project.PharaohProject = self.sphinx_app.pharaoh_proj

        if isinstance(docname, str):
            template_file = Path(self.sphinx_app.env.doc2path(docname))
        elif isinstance(docname, Path):
            template_file = docname
        else:
            msg = f"Unsupported type ({type(docname)!r}) for argument 'docname'!"
            raise TypeError(msg)

        try:
            # Find component name by parent directory
            components_dir = project.sphinx_report_project_components
            try:
                parts = template_file.relative_to(components_dir).parts
                component_name = "" if len(parts) < 2 else parts[0]
            except ValueError:
                component_name = ""

            # Context Creation
            log.debug(f"Discovering additional templating context for component {component_name!r}...")
            self.default_context["project"]["component_name"] = component_name
            self.default_context["local"] = {}

            for key, ctx in self.read_local_context_files(self.default_context, template_file.parent).items():
                log.debug(f"... discovered local context {key!r}")
                if key in self.default_context["local"]:
                    log.warning(f"Overwriting existing local context namespace {key!r}!")
                self.default_context["local"][key] = ctx

            for key, ctx in self.read_asset_context_files(component=component_name):
                log.debug(f"... discovered asset context {key!r}")
                if key in self.default_context["local"]:
                    log.warning(f"Overwriting existing local context namespace {key!r}!")
                self.default_context["local"][key] = ctx

            with chdir(template_file.parent):
                template = self.select_template([template_file.name], parent=str(template_file.parent))
                rendered = template.render(
                    {"ctx": self.default_context}, **self.get_render_globals(project, component_name, template_file)
                )

            rendered_file = template_file.parent / (template_file.name + ".rendered")
            rendered_file.write_text(rendered)
            return rendered, rendered_file
        except Exception:
            log.error(
                f"Error in PharaohTemplateEnv.sphinx_source_read_hook while trying to render template "
                f"{template_file} with context {pprint.pformat(self.default_context)}!",
                exc_info=True,
            )
            raise
        finally:
            self.default_context["local"] = {}

    def get_render_globals(
        self, project: pharaoh.project.PharaohProject, component_name: str, template_file: Path
    ) -> dict[str, Callable]:
        return {
            "search_error_assets": functools.partial(
                project.asset_finder.search_assets,
                components=[component_name],
                condition="asset_type == 'error_traceback'",
            ),
            "search_assets": functools.partial(project.asset_finder.search_assets, components=[component_name]),
            "asset_rel_path_from_project": partial(asset_rel_path_from_project, project),
            "asset_rel_path_from_build": partial(asset_rel_path_from_build, self.sphinx_app, template_file),
        }

    def join_path(self, template: str, parent: str) -> str:
        """
        Join a template with the parent. By default, all the lookups are
        relative to the loader root so this method returns the `template`
        parameter unchanged, but if the paths should be relative to the
        parent template, this function can be used to calculate the real
        template name.

        Subclasses may override this method and implement template path
        joining here.
        """
        if "<>" in parent:
            # This case happens only when pharaoh is deployed at site-packages (don't know why, no time to find out).
            # In this case under Python 3.7, the following parent.is_dir() is failing because os.stat
            # does not allow <> in the file path, Python 3.9 does, apparently.
            # So we split off the suffix after <> (which would be done anyway using the "parent.parent" statement).
            parent = parent.split("<>")[0]
        parent_path = Path(parent)
        if not parent_path.is_dir():
            parent_path = parent_path.parent
        # Use a special illegal string <> to be able to split the template from the parent in
        # PharaohFileSystemLoader.get_source
        return parent_path.as_posix() + "<>" + Path(template).as_posix()

    def read_local_context_files(self, base_context: dict, basepath: Path) -> dict:
        """
        Reads all *_context.yaml and executes all *_context.py files to collect additional context data for templating.
        """
        local_context: dict = {}
        for file in basepath.glob("*_context.yaml"):
            f = Path(file)
            key = f.name.rsplit("_", 1)[0]
            with f.open():
                conf = omegaconf.OmegaConf.load(f)
                omegaconf.OmegaConf.resolve(conf)
                local_context[key] = conf

        for file in basepath.glob("*_context.py"):
            f = Path(file)
            key = f.name.rsplit("_", 1)[0]
            if f in self.local_context_file_cache:
                module = self.local_context_file_cache[f]
            else:
                module = module_from_file(f)
                run_module(module, f.read_text(encoding="utf-8"))
                self.local_context_file_cache[f] = module
            result_ctx: dict = module.__dict__
            if not ("context" in result_ctx and isinstance(result_ctx["context"], dict)):
                msg = (
                    "No dict-typed variable named 'context' was defined by the script!\n"
                    "Example of valid script content: context={}"
                )
                raise LookupError(msg)
            if key in local_context:
                log.warning(f"Local context key {key!r} already exists and will be overwritten by {file}")
            local_context[key] = result_ctx["context"]

        return local_context

    def read_asset_context_files(self, component: str) -> Iterator[tuple[str, list | dict]]:
        proj = self.sphinx_app.pharaoh_proj
        for asset in proj.asset_finder.search_assets(
            components=[component], condition="pharaoh_templating_context != ''"
        ):
            suffix = asset.assetfile.suffix.lower()
            if suffix == ".json":
                context = asset.read_json()
            elif suffix == ".yaml":
                context = asset.read_yaml()
            else:
                msg = f"Unsupported file suffix {suffix!r}"
                raise NotImplementedError(msg)
            yield asset.context.pharaoh_templating_context, context


def module_from_file(path: str | Path) -> ModuleType:
    """
    Creates a module from file at runtime.

    :param path: Path to the module source code.
    :return: A fresh module
    """

    module_path = Path(path)
    module_name = f"local_context_{module_path.stem.replace(' ', '_').replace('-', '_')}_{uuid.uuid4().hex[:6]}"
    module = ModuleType(module_name)
    module.__dict__["__file__"] = str(module_path.absolute())
    module.__path__ = [str(module_path.parent)]
    module.__dict__["__name__"] = "__main__"
    module.__dict__["__module_name__"] = module_name

    return module


def run_module(module: ModuleType, code: str):
    """
    Execute the configured source code in a module.

    :param module: A module object.
    :param code: The code to be run inside the module
    """
    compiled_code = compile(code, module.__dict__["__file__"], "exec")

    target_wd = Path(module.__dict__["__file__"]).parent
    with chdir(target_wd):
        exec(compiled_code, module.__dict__)



from __future__ import annotations

import json
from functools import singledispatch
from pathlib import Path

try:
    import numpy as np

    NUMPY_AVAIL = True
except ImportError:
    NUMPY_AVAIL = False


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        for type_, handler in encode.registry.items():
            if isinstance(obj, type_) and type_ is not object:
                return handler(obj)
        try:
            return super().default(obj)
        except Exception as e:
            print(f"CustomJSONEncoder Error: {e} when trying to encode {obj!r}")
            raise


@singledispatch
def encode(obj, **kwargs):
    return json.dumps(obj, cls=CustomJSONEncoder, **kwargs)


@encode.register(Path)
def encode_Path(obj):
    """Event-typed object cannot be JSON serialized, so serialize as "null" """
    return str(obj)


if NUMPY_AVAIL:

    @encode.register(np.str_)
    @encode.register(np.bytes_)
    @encode.register(np.bool_)
    @encode.register(np.number)
    def encode_numpy_scalars(obj):
        return obj.item()

    @encode.register(np.ndarray)
    def encode_numpy_arrays(obj):
        return obj.tolist()


def encode_dict(obj, **kwargs):
    return json.loads(encode_json(obj, **kwargs))


def encode_json(obj, **kwargs):
    return json.dumps(obj, cls=CustomJSONEncoder, **kwargs)



from __future__ import annotations

from contextlib import ExitStack, contextmanager


@contextmanager
def patch_3rd_party_libraries():
    """
    A context manager that patches certain save/show functions of the supported libraries (mostly plotting frameworks).
    When exiting this context the patching will be undone.

    show/display functions become no-ops and functions outputting files are patching the storage location
    to be inside <pharaoh-project>/pharaoh_assets/build. For interactive plots the options that control things like
    js/css resources are patched to use a single copy inside the asset folder to avoid wasting diskspace for each plot.
    For static images

    User perspective:
        The patched functions only change their behavior when the script is called by pharaoh. When the script is called
        standalone then the behavior remains unchanged.

    .. important::

        When using this function, either use it as a context manager or instantiate it, assign it to a variable and
        call ``__enter()__`` on it. If no reference to the context manager is kept, it gets garbage-collected and
        shuts down, which leads to a premature reset of patches.

    """
    from pharaoh.assetlib.patches import _bokeh, _holoviews, _matplotlib, _pandas, _panel, _plotly

    patch_modules = (_bokeh, _holoviews, _matplotlib, _pandas, _plotly, _panel)

    with ExitStack() as stack:
        for module in patch_modules:
            if hasattr(module, "init_module"):
                module.init_module()
            stack.enter_context(module.patch())
        try:
            yield
        finally:
            # In error case, close the exit stack explicitly here,
            # since yield will raise an exception if the asset script execution results in an error.
            # So we have to handle this error and undo the patching.
            # Previously there was no try...except around the yield, which led to not un-patching
            # the 3rd-party packages.
            stack.close()



from __future__ import annotations

import datetime
import io
import os
import typing as t
from pathlib import Path

from pharaoh.log import log

if t.TYPE_CHECKING:
    import matlab.engine as me


def ensure_engine():
    try:
        import matlab.engine  # noqa: F401
    except Exception:
        log.error(
            "The Matlab Engine for Python is not installed.\n"
            "Please install the corresponding version via PIP:\n"
            "  R2020B: pip install matlabengine==9.9.*\n"
            "  R2021A: pip install matlabengine==9.10.*\n"
            "  R2021B: pip install matlabengine==9.11.*\n"
            "  R2022A: pip install matlabengine==9.12.*\n"
            "  R2022B: pip install matlabengine==9.13.*\n"
            "  R2023A: pip install matlabengine==9.14.*\n"
            "  R2023B: pip install matlabengine==9.15.*\n"
        )
        raise


PathLike = t.Union[str, Path]


class Matlab:
    """
    Matlab engine for Pharaoh asset generation.

    Usage::

        from pharaoh.assetlib.api import Matlab

        eng = Matlab()
        with eng:
            out, err = eng.execute_script("myscript.m")
            result, out, err = eng.execute_function("myfunc", [800.0], nargout=1)

    """

    def __init__(self, start_options: str = "-nodesktop"):
        """
        Matlab engine for Pharaoh asset generation.

        :param start_options: See options at https://de.mathworks.com/help/matlab/ref/matlabwindows.html
        """
        self._engine: me.MatlabEngine | None = None
        self._start_options = start_options

    @property
    def eng(self) -> me.MatlabEngine:
        """
        Returns the matlab engine instance
        """
        return self._engine

    def __enter__(self):
        """
        Context manager enter. Connects to Matlab.
        """
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager exit. Disconnects from Matlab.
        """
        if self.eng is not None:
            self.eng.exit()
            self._engine = None

    def connect(self):
        """
        Connects to a Matlab engine. If an engine is running it will be connected, otherwise a new Matlab instance
        will be started and connected.
        """
        ensure_engine()
        import matlab.engine as me

        active_engines = me.find_matlab()
        if len(active_engines):
            self._engine = me.connect_matlab(name=active_engines[0], background=False)
        else:
            self._engine = me.start_matlab(option=self._start_options, background=False)

        # Write connection info to Matlab workspace
        info = datetime.datetime.now(tz=datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S") + f", PID {os.getpid()}"
        self.eng.workspace["PharaohConnected"] = info

    def show_gui(self):
        """
        Makes the Matlab GUI visible.
        """
        self.eng.desktop(nargout=0)

    def execute_script(self, script_path: PathLike) -> tuple[str, str]:
        """
        Executes a Matlab script.

        Returns a tuple of strings containing the stdout and stderr streams of the script execution.
        """
        script_path = Path(script_path).absolute()

        # Call the script
        fname = script_path.stem
        workdir = str(script_path.parent) if script_path.exists() else None
        _, out, err = self.execute_function(function_name=fname, args=None, nargout=0, workdir=workdir)
        return out, err

    def execute_function(
        self,
        function_name: str,
        args: list[t.Any] | None = None,
        nargout: int = 0,
        workdir: PathLike | None = None,
    ) -> tuple[t.Any, str, str]:
        """
        Executes a Matlab function

        Returns a tuple of result, stdout stream and stderr stream of the function execution.

        The shape/type of result depends on the argument 'nargout'. See below.

        :param function_name:
            The name of the function to execute.

            The function must be in the Matlab path to execute it.

            Alternatively the workdir argument can be set to the parent directory of the function.
        :param args: A list of positional input arguments to the function
        :param nargout:
            The number of output arguments.

            If 0, the result return value of this function will be None.

            If 1, the result return value will be a single value.

            If greater than 1, the result return value will be a tuple containing nargout values.
        :param workdir: The Matlab working directory to be changed to during function execution. Skipped if None.
        """
        old_cwd = self.eng.cd(str(workdir)) if workdir is not None else None
        try:
            out = io.StringIO()
            err = io.StringIO()
            result = getattr(self.eng, function_name)(
                *(args or []), background=False, nargout=max(0, int(nargout)), stdout=out, stderr=err
            )
            return result, out.getvalue(), err.getvalue()
        except Exception as e:
            log.error(f"Execution of Matlab function/script {function_name!r} failed: {e}", exc_info=True)
            raise
        finally:
            if old_cwd:
                self.eng.cd(old_cwd, background=False)



from __future__ import annotations

import getpass
import logging
import os
import platform
import re
import sys
from pathlib import Path

import pharaoh

log = logging.getLogger("pharaoh")
log.setLevel(logging.WARNING)


FMT = logging.Formatter("%(levelname)-7s|%(asctime)s.%(msecs)03d|  %(message)s")


def add_streamhandler():
    h = logging.StreamHandler(sys.stdout)
    h.setFormatter(FMT)
    h.name = "pharaoh_sh"
    for hdl in log.handlers:
        if hdl.name == h.name:
            log.removeHandler(hdl)
            break
    log.addHandler(h)


def remove_filehandlers(path: Path):
    for hdl in log.handlers[:]:
        if isinstance(hdl, logging.FileHandler):
            try:
                Path(hdl.baseFilename).relative_to(path)
                log.removeHandler(hdl)
            except ValueError:
                pass


def add_filehandler(path: Path):
    h = logging.FileHandler(path / "log.txt", mode="a", delay=True)
    h.setFormatter(FMT)
    h.name = "pharaoh_fh"
    for hdl in log.handlers:
        if hdl.name == h.name:
            log.removeHandler(hdl)
            break
    log.addHandler(h)


def add_warning_filehandler(path: Path):
    h = logging.FileHandler(path / "log_warnings.txt", mode="a", delay=True)
    h.setFormatter(FMT)
    h.setLevel(logging.WARNING)
    h.name = "pharaoh_fh_warn"
    for hdl in log.handlers:
        if hdl.name == h.name:
            log.removeHandler(hdl)
            break
    log.addHandler(h)


class SphinxLogRedirector:
    ansi_re = re.compile("\x1b\\[(\\d\\d;){0,2}\\d\\dm")

    def __init__(self, level: int):
        self.log_func = {
            logging.DEBUG: log.debug,
            logging.INFO: log.info,
            logging.WARNING: log.warning,
        }[level]

    def write(self, data: str):
        line = self.ansi_re.sub("", data).rstrip()
        if line:
            self.log_func(f"Sphinx Build | {line}")

    def flush(self):
        pass


def log_version_info():
    level = log.level
    log.setLevel(logging.INFO)
    try:
        log.info(f"Pharaoh {pharaoh.__version__} [Python {platform.python_version()}, {platform.platform()}]")
    finally:
        log.setLevel(level)


def log_debug_info():
    envs = dict(os.environ)
    envs.pop("LIBRARY_ROOTS", None)
    envs.pop("_OLD_VIRTUAL_PATH", None)
    level = log.level
    log.setLevel(logging.INFO)
    try:
        log.info(f"Pharaoh version: {pharaoh.__version__}")
        log.info(f"Python Version:  {platform.python_version()}")
        log.info(f"Python Binary:   {sys.executable}")
        log.info(f"System:          {platform.platform()}")
        log.info(f"User:            {getpass.getuser()}")
        log.info(f"Path:            {os.environ.get('PATH', '')}")
        log.info(f"PythonPath:      {os.environ.get('PYTHONPATH', '')}")
    finally:
        log.setLevel(level)


if __name__ == "__main__":
    add_streamhandler()
    log_debug_info()



# mypy: disable-error-code="method-assign"
from __future__ import annotations

import os
from contextlib import contextmanager

try:
    import matplotlib.pyplot as plt
    from matplotlib.figure import Figure as mpl_figure

    MPL_AVAIL = True
except (ImportError, ModuleNotFoundError):
    MPL_AVAIL = False

from pharaoh import project
from pharaoh.assetlib.context import context_stack
from pharaoh.assetlib.util import parse_signature

if MPL_AVAIL:
    # memorize the unpatched functions
    vanilla_mpl_figure_savefig = mpl_figure.savefig
    vanilla_mpl_figure_show = mpl_figure.show
    vanilla_mpl_plt_show = plt.show

    @contextmanager
    def patch():
        """
        A context manager that patches all functions of matplotlib relevant to display/save figures when executed in a
        pharaoh context. When the program exits the patch-context, the patching will be undone.

        The main goals:
            - retain original function call signature
            - suppress display of figures in GUI or browser
            - alter storage location of figures to the pharaoh build folder
            - inject static image parameters via config file
        """

        # switching out the original with the patched functions
        if mpl_figure.savefig == vanilla_mpl_figure_savefig:
            mpl_figure.savefig = patched_savefig
            mpl_figure.show = patched_mpl_figure_show
            plt.show = patched_mpl_plt_show

        yield

        # undo the patching
        if mpl_figure.savefig != vanilla_mpl_figure_savefig:
            mpl_figure.savefig = vanilla_mpl_figure_savefig
            mpl_figure.show = vanilla_mpl_figure_show
            plt.show = vanilla_mpl_plt_show

    def patched_mpl_figure_show(*args, **kwargs):
        """
        This is the patched version of ``matplotlib.figure.Figure.show(warn=True)``.

        Changes:
            - becomes a no-op when executed in a pharaoh context

        docs: https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure.show
        """

    def patched_mpl_plt_show(*args, **kwargs):
        """
        This is the patched version of ``matplotlib.pyplot.show(*, block=None)``.

        Changes:
            - becomes a no-op when executed in a pharaoh context

        docs: https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.show.html
        """

    def patched_savefig(*args, **kwargs):
        """
        This is the patched version of ``matplotlib.figure.Figure.savefig(fname, *, transparent=None, **kwargs)``.

        Changes:
            - takes kwargs from the pharaoh config file and updates the kwargs passed into the function
            - changes the plots storage destination to be inside pharaoh's asset build folder

        docs: https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure.savefig
        """
        params = parse_signature(obj=vanilla_mpl_figure_savefig, args=args, kwargs=kwargs, fromclass=True)
        params.pop("kwargs")
        params.update(kwargs)
        filename = params.pop("fname")

        format: str | None = params.pop("format", None)  # Format must be given through filename
        if format is None:
            # get format from filename, or from backend's default filetype
            if isinstance(filename, os.PathLike):
                filename = os.fspath(filename)
            if isinstance(filename, str):
                format = os.path.splitext(filename)[1][1:]
            if format is None or format == "":
                format = plt.FigureCanvasBase.get_default_filetype()
                if isinstance(filename, str):
                    filename = filename.rstrip(".") + "." + format.lower()
        else:
            format = format.lower().lstrip(".")
            filename = os.path.splitext(filename)[0] + "." + format

        active_app = project.get_project()
        params.update(active_app.get_setting("toolkits.matplotlib.savefig"))

        if format in ("jpeg", "jpg", "png", "svg"):
            template_context = {"template": "image"}
        else:
            msg = f"File format of {filename!r} not supported!"
            raise NotImplementedError(msg)

        file_path = active_app._build_asset_filepath(filename)
        params["filename"] = str(file_path)
        with context_stack.new_context(
            asset=dict(
                user_filepath=str(filename),
                file=str(file_path),
                name=file_path.name,
                stem=file_path.stem,
                suffix=file_path.suffix,
                **template_context,
            )
        ):
            context_stack.get_parent_context(name="generate_assets")["asset"]["index"] += 1
            vanilla_mpl_figure_savefig(params.pop("self"), params.pop("filename"), **params)
            context_stack.dump(file_path)

else:
    # if the package is not installed there is nothing to patch, so we silently do nothing when patch() is called.

    @contextmanager
    def patch():
        yield



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pharaoh.templating.second_level.template_env import PharaohTemplateEnv


def find_asset_template(template_name: str) -> Path:
    from pharaoh.plugins.plugin_manager import PM

    files = PM.pharaoh_find_asset_render_template(template_name)
    if not files:
        msg = f"Cannot find template that matches {template_name!r}!"
        raise LookupError(msg)
    if len(files) > 1:
        files_listed = "\n  ".join(map(str, files))
        msg = f"Found multiple matches for {template_name!r}:\n  {files_listed}\nPlease use a more specific name!"
        raise LookupError(msg)
    return files[0]


def render_asset_template(jinja_env: PharaohTemplateEnv, template: str, **kwargs) -> str:
    template_file = Path(template).absolute() if Path(template).absolute().exists() else find_asset_template(template)
    template_content = "\n" + template_file.read_text(encoding="utf-8").strip() + "\n"

    tmpl = jinja_env.from_string(template_content)
    return tmpl.render(**kwargs)



from __future__ import annotations

from pathlib import Path


def find_template(template: str) -> Path:
    from pharaoh.plugins.plugin_manager import PM

    templates = PM.pharaoh_collect_l1_templates()
    if isinstance(template, str):
        if template in templates:
            return Path(templates[template].path).absolute()
        ptemplate = Path(template)
        if ptemplate.exists() and str(ptemplate) not in ("", "."):
            return Path(template)

        msg = f"Not a valid template {template!r}! Available templates: {', '.join(templates)}."
        raise ValueError(msg)

    msg = f"Not a valid template {template!r}! Available templates: {', '.join(templates)}."
    raise ValueError(msg)



from __future__ import annotations

import datetime
import functools
import logging
import os
import re
import shutil
import sys
import traceback
import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

import attr
import attrs
import omegaconf
import omegaconf.errors

import pharaoh
import pharaoh.log
import pharaoh.util.oc_resolvers
from pharaoh.assetlib import finder, resource
from pharaoh.assetlib.context import context_stack
from pharaoh.errors import AssetGenerationError, ProjectInconsistentError
from pharaoh.plugins.plugin_manager import PM
from pharaoh.util.contextlib_chdir import chdir

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    import jinja2

try:
    PHARAOH_CLI_PATH = Path(shutil.which("pharaoh.exe")).as_posix()  # type: ignore[arg-type]
except TypeError:
    PHARAOH_CLI_PATH = "pharaoh.exe"

PathLike = Union[str, Path]

LRU_PHARAOH_PROJECT: PharaohProject | None = None

DEFAULT_MISSING = object()

log = pharaoh.log.log


def get_project(lookup_path: str | Path | None = None) -> PharaohProject:
    """
    Returns an instance of PharaohProject.

    :param lookup_path: If None, the function tries to return the Pharaoh singleton instance that is already loaded in
                        memory and raises and exception if there is none.
                        If given a string or Path instance, searches the given path and all its parent folders
                        for a ``pharaoh.yaml`` project file and returns a PharaohProject instance.
    """
    if lookup_path is None:
        # Enforce having a singleton per process
        if LRU_PHARAOH_PROJECT is None:
            msg = "No PharaohProject instance created yet!"
            raise RuntimeError(msg)
        return LRU_PHARAOH_PROJECT

    if isinstance(lookup_path, (str, Path)):
        lookup_path = Path(lookup_path)
        if lookup_path.is_file():
            lookup_path = lookup_path.parent
    else:
        msg = "Unsupported value for argument 'lookup_path'."
        raise ValueError(msg)

    path = Path(lookup_path).absolute()
    if path.is_file():
        path = path.parent
    while True:
        if (path / PharaohProject.PROJECT_SETTINGS_YAML).exists():
            return PharaohProject(project_root=path)

        if path == path.parent:
            # We reached the path's anchor level
            msg = (
                f"Cannot find a Pharaoh project file {PharaohProject.PROJECT_SETTINGS_YAML} in "
                f"{lookup_path} or one of its parent directories!"
            )
            raise LookupError(msg)

        path = path.parent


class PharaohProject:
    PROJECT_SETTINGS_YAML = "pharaoh.yaml"

    def __new__(cls, *args, **kwargs):
        # Cache the last recently used project instance in a global variable,
        # so it can be accessed from anywhere after creation
        global LRU_PHARAOH_PROJECT  # noqa: PLW0603
        LRU_PHARAOH_PROJECT = object.__new__(cls)
        return LRU_PHARAOH_PROJECT

    def __del__(self):
        try:  # noqa: SIM105
            pharaoh.log.remove_filehandlers(self._project_root)
        except Exception:
            pass

    def __init__(
        self,
        project_root: PathLike,
        overwrite: bool = False,
        templates: str | list[str] | tuple[str] = ("pharaoh.default_project",),
        template_context: dict[str, Any] | None = None,
        **kwargs,
    ):
        """
        Instantiates a Pharaoh project instance using either an existing project root
        (directory containing pharaoh.yaml) or creates a new project.

        :param project_root: A directory containing a pharaoh.yaml file
        :param overwrite: If the project directory already contains files, overwrite them.
        :param templates: The project template(s) to use for creating the Sphinx project files.
                          Maybe be any number of template names or paths to directories.
        :param template_context: The Jinja rendering context for the selected project templates.
        :keyword custom_settings: A path to a YAML file containing settings to overwrite the default project settings.
        """
        self._settings_map: dict[str, omegaconf.DictConfig] = {}
        self._merged_settings: omegaconf.DictConfig = omegaconf.DictConfig({})
        self._project_root: Path = Path(project_root).absolute().resolve()
        self._asset_finder: finder.AssetFinder | None = None

        logging_add_filehandler = kwargs.pop("logging_add_filehandler", True)
        custom_settings = kwargs.pop("custom_settings", None)
        if kwargs:
            msg = f"Unknown keyword arguments {tuple(kwargs.keys())}!"
            raise ValueError(msg)

        pharaoh.log.add_streamhandler()
        if logging_add_filehandler:
            pharaoh.log.add_filehandler(self._project_root)
            pharaoh.log.add_warning_filehandler(self._project_root)

        if isinstance(templates, str):
            templates = [templates]
        self._ensure_project(
            templates=list(templates),
            template_context=template_context,
            recreate=overwrite,
            custom_settings=custom_settings,
        )

        self.load_settings(namespace="all")
        log.setLevel(self.get_setting("logging.level").upper())

    def load_settings(self, namespace: str = "all"):
        """
        Loads settings from various namespaces.

        default
            The default setting in the Pharaoh library

        project
            The project settings that may be modified by the user

        env
            The settings defined by environment variables starting with PHARAO

        all
            Loads all of the above

        :param namespace: The namespace to load settings from. all, default, project or env.
        """
        if namespace in ("all", "default"):
            self._settings_map["default"] = self._load_default_settings()

        if namespace in ("all", "project"):
            project_settings = self._project_root / self.PROJECT_SETTINGS_YAML
            with open(project_settings) as fp:
                self._settings_map["project"] = omegaconf.OmegaConf.load(fp)  # type: ignore[assignment]

        if namespace in ("all", "env"):
            dotlist = []
            env_pat = re.compile(r"^PHARAOH?((\.)|(__)).+$", re.IGNORECASE)
            for k, v in os.environ.items():
                if env_pat.fullmatch(k):
                    keys = re.split(r"\.|(?<!_)__(?!_)", k)
                    key = ".".join(keys[1:]).lower()
                    dotlist.append(f"{key}={v}")
            if dotlist:
                self._settings_map["env"] = omegaconf.OmegaConf.from_dotlist(dotlist)
            else:
                self._settings_map["env"] = omegaconf.OmegaConf.create({})
        self._update_merged_settings()

    def _update_merged_settings(self):
        """
        Merges settings from all namespaces.
        """
        self._merged_settings = omegaconf.OmegaConf.create(
            omegaconf.OmegaConf.merge(
                self._settings_map["default"],
                self._settings_map["project"],
                self._settings_map["env"],
            )
        )  # type: ignore[assignment]

    def get_settings(self) -> omegaconf.DictConfig:
        """
        Returns merged settings from all namespaces.
        """
        if self._merged_settings is None:
            self._update_merged_settings()
        return self._merged_settings

    def put_setting(self, key: str, value: Any):
        """
        Sets a setting by its dot-separated name, e.g. "core.debug".

        This has no effect, if there is an environment variable defining the same setting!
        """
        # todo: Let put_setting be able to overwrite settings defined in env vars.
        #       Currently env vars have priority.
        paths = key.lower().split(".")
        node = self._project_settings
        for path in paths[:-1]:
            node = getattr(node, path)
        setattr(node, paths[-1], value)
        self._update_merged_settings()

    def get_setting(self, key: str, default=DEFAULT_MISSING, to_container: bool = False, resolve: bool = True):
        """
        Gets a setting by its dot-separated name, e.g. "core.debug".

        The settings are preferably taken from environment variables (e.g. PHARAOH.CORE.DEBUG), whereas the values
        must be TOML compatible values (e.g. "true" for boolean).

        If no environment variable is found, the user defined settings in the pharaoh.yaml file in the
        project root will be used. If the setting is not specified in there or the file is not existing,
        the Pharaoh default settings are queried.

        If the setting is not found, a LookupError is returned unless "default" is set.

        Examples::

            proj.get_setting("report.title")
            proj.get_setting("cloud.metadata", to_container=True, resolve=False) == {
                "author": "${report.author}",
                "title": "${report.title}",
            }
            proj.get_setting("cloud.metadata", to_container=True, resolve=True) == {
                "author": "loibljoh",
                "title": "Pharaoh Report",
            }
            proj.get_setting("cloud.metadata.title") == "Pharaoh Report"
            proj.get_setting("cloud.metadata") == omegaconf.dictconfig.DictConfig(...)

        :param key: Dot-separated name of the setting value to return
        :param default: The returned default value if the key does not exist.
        :param to_container: If set, recursively converts an OmegaConf config to a primitive Python container type
            (dict or list).
        :param resolve: True to resolve all values if "to_container" is set to True
        """
        if not key:
            msg = "Must be a string with a minimum length of 1!"
            raise ValueError(msg)
        merged_settings = self.get_settings()
        key = key.lower()
        ret = omegaconf.OmegaConf.to_container(merged_settings, resolve=resolve) if to_container else merged_settings
        try:
            paths = key.split(".")
            for path in paths:
                ret = ret[path]

            return ret
        except KeyError:
            if default is DEFAULT_MISSING:
                msg = f"No setting key called {key!r}!"
                raise LookupError(msg) from None
            return default

    def save_settings(self, include_env: bool = False):
        """
        Saves back the project settings to the project YAML file.

        :param include_env: If True, Pharaoh settings that are set via environment variables will be persisted
            to the project settings YAML file.
        """
        project_settings = self._project_root / self.PROJECT_SETTINGS_YAML
        if include_env:
            settings = omegaconf.OmegaConf.merge(
                self._settings_map["project"],
                self._settings_map["env"],
            )
        else:
            settings = self._project_settings
        with open(project_settings, "w") as fp:
            omegaconf.OmegaConf.save(config=settings, f=fp)

    @property
    def _project_settings(self):
        return self._settings_map["project"]

    @property
    def settings_file(self):
        return self._project_root / self.PROJECT_SETTINGS_YAML

    @property
    def project_root(self):
        return self._project_root

    @property
    def sphinx_report_project(self):
        return self.project_root / "report-project"

    @property
    def sphinx_report_project_components(self):
        return self.sphinx_report_project / "components"

    @property
    def sphinx_report_build(self):
        return self.project_root / "report-build"

    @property
    def asset_build_dir(self):
        return self.sphinx_report_project / ".asset_build"

    @property
    def asset_finder(self) -> finder.AssetFinder:
        if self._asset_finder is None:
            self._asset_finder = finder.AssetFinder(self.asset_build_dir)
        return self._asset_finder

    def add_component(
        self,
        component_name: str,
        templates: str | (Path | (Iterable[str] | Iterable[Path])) = ("pharaoh.empty",),
        render_context: dict | None = None,
        resources: list[resource.Resource] | None = None,
        metadata: dict[str, Any] | None = None,
        index: int = -1,
        overwrite: bool = False,
    ):
        """
        Adds a new component to the Pharaoh project

        Example::

            from pharaoh.api import FileResource, PharaohProject

            proj = PharaohProject(".")
            proj.add_component(
                component_name="component_ABC",
                templates=[
                    "plugin_abc.template_xyz",              # plugin template
                    "path/to/template/directory",           # template directory
                    "path/to/template/file/tmpl.pharaoh.py" # template file
                ],
                render_context={"foo": "bar"},
                resources=[FileResource(alias="dlh5_result", pattern="C:/temp/**/*.dlh5")],
                metadata={"some tag": "some value"},
                index=-1,  # append
                overwrite=False
            )

        :param component_name: The name of the component. Must be a valid Python identifier.
        :param templates: A list of component templates to use for creating the components project files.
            Those may be the template identifier (e.g. ``plugin_abc.template_xyz``) of a registered plugin template,
            a path to a template directory or a path to a template file (single-file template).

            Since multiple template may be specified, their order matters in cases where different templates
            create equally-named files, thus templates might overwrite files of the previous templates.

            This enables *template-composition*, where template designers can chunk their bigger templates into
            smaller reusable building blocks.

            If omitted, an empty default template is used.

        :param render_context: The Jinja rendering context for the selected template.
                               The actual template rendering context will have ``component_name``, ``resources`` and
                               ``metadata`` available under the respective keys.
        :param resources: A list of Resource instances, defining the component's resources used in asset scripts
        :param metadata: A dictionary of metadata that may be used to find the component via
                         :func:`PharaohProject.find_component()
                         <pharaoh.project.PharaohProject.find_component>` method.
        :param overwrite: If True, an already existing component will be overwritten
        """
        from pharaoh.templating.first_level.find import find_template
        from pharaoh.templating.first_level.render import render_template

        # Check if component with this name already added
        components = []
        for component in self.iter_components():
            if component["name"] == component_name:
                if overwrite:
                    pass
                else:
                    msg = f"Component {component_name!r} already exists!"
                    raise KeyError(msg)
            else:
                components.append(component)

        render_context = render_context or {}
        render_context["pharaoh_cli_path"] = PHARAOH_CLI_PATH

        if isinstance(templates, (str, Path)):
            templates = [str(templates)]
        templates = [str(t) for t in templates]

        if not templates:
            msg = "No templates specified!"
            raise ValueError(msg)

        new_component = Component(
            name=component_name,
            templates=templates,
            render_context=render_context,
            resources=resources or [],
            metadata=metadata or {},
        )

        # Support negative indexing where -1 is the last item. Normal insert with -1 will insert at second last index.
        if index < 0:
            index = max(0, len(components) + 1 + index)

        components.insert(
            index,
            attrs.asdict(new_component),  # type: ignore[arg-type]
        )

        for templ in templates:
            template_path = find_template(templ)
            if (template_path / "[[ component_name ]]").exists():  # pragma: no cover
                msg = (
                    "The usage of [[ component_name ]] as top level directory name of component "
                    "templates is not required anymore. "
                )
                raise RuntimeError(msg)

            render_template(
                template_path=template_path,
                outputdir=self.sphinx_report_project_components / component_name,
                context=new_component.get_render_context(),
            )

        self._project_settings.components = components
        self.save_settings()
        log.info(f"Added component {component_name!r}. Saved project.")

    def add_template_to_component(
        self, component_name: str, templates: str | Path | Iterable[str], render_context: dict | None = None
    ):
        """
        Adds additional templates to an existing component, that may overwrite existing files during rendering.

        :param component_name: The name of the component. Must be a valid Python identifier.
        :param templates: The component template(s) to use for creating the components project files.
                          Maybe be the a single or multiple template names or paths to a directory.
        :param render_context: The Jinja rendering context for the selected template.
                               The actual template rendering context will have component_name, resources and metadata
                               available under the respective keys.
        """
        from pharaoh.templating.first_level.find import find_template
        from pharaoh.templating.first_level.render import render_template

        # Check if component with this name already added
        for component_index, comp in enumerate(self.iter_components()):  # noqa: B007
            if comp["name"] == component_name:
                component = Component.from_dictconfig(comp)
                break
        else:
            msg = f"Component {component_name!r} does not exist!"
            raise KeyError(msg)

        component.render_context.update(render_context or {})

        if isinstance(templates, (str, Path)):
            templates = [str(templates)]
        templates = [str(t) for t in templates]

        if not templates:
            msg = "No templates specified!"
            raise ValueError(msg)

        component.templates.extend(templates)

        for templ in templates:
            template_path = find_template(templ)
            render_template(
                template_path=template_path,
                outputdir=self.sphinx_report_project_components / component_name,
                context=component.get_render_context(),
            )

        self._project_settings.components[component_index] = attrs.asdict(component)

        self._update_merged_settings()
        self.save_settings()
        log.info(f"Updated component {component_name!r}. Saved project.")

    def remove_component(
        self,
        filter: str,
        regex: bool = False,
    ) -> list[str]:
        """
        Removes one or multiple existing components.

        :param filter: A case-insensitive component filter. Either a full-match or regular expression, depending on
            regex argument.
        :param regex: If True, the filter argument will be treated as regular expression. Components
            that partially match the regular expression are removed.
        :returns: A list of component names that got removed
        """
        # Check if component with this name already added
        if regex:
            rex = re.compile(filter, flags=re.IGNORECASE)
        else:
            filter = filter.lower()
        removed = []
        components = []
        for comp in self.iter_components():
            if (regex and rex.match(comp["name"]) is not None) or comp["name"].lower() == filter:
                comp_files = self.sphinx_report_project_components / comp["name"]
                if comp_files.exists() and comp_files.is_dir():  # pragma: no cover
                    shutil.rmtree(comp_files)
                removed.append(comp["name"])
            else:
                components.append(comp)

        if removed:
            self._project_settings.components = components
            self.save_settings()
            log.info(f"Removed components {','.join(removed)}. Saved project.")
        return removed

    def iter_components(self, filtered: bool = False) -> Iterator[omegaconf.DictConfig]:
        """
        Returns an iterator over all components from a project.
        """
        filter_include = self.get_setting("report.component_filter.include", ".*")
        filter_exclude = self.get_setting("report.component_filter.exclude", None)
        components = self._project_settings.get("components", []) or []

        # input validation
        filter_include = ".*" if filter_include is None else filter_include
        if not isinstance(filter_include, str):
            msg = "Invalid pharaoh.yaml: report.component_filter.include must be either set to a string or left empty."
            raise ValueError(msg)
        if not (filter_exclude is None or isinstance(filter_exclude, str)):
            msg = "Invalid pharaoh.yaml: report.component_filter.exclude must be either set to a string or left empty."
            raise ValueError(msg)

        if filtered:
            for component in components:
                filters_matched = False
                if re.fullmatch(filter_include, component.name) is not None and (
                    filter_exclude is None or re.fullmatch(filter_exclude, component.name) is None
                ):
                    filters_matched = True
                if filters_matched:
                    yield component
        else:
            yield from components

    def find_components(self, expression: str = "", filtered: bool = False) -> list[omegaconf.DictConfig]:
        """
        Find components by their metadata using an evaluated expression.

        The expression must be a valid Python expression and following local variables may be used:

        - name: The name of the component
        - templates: A list of templates used to render the component
        - metadata: A dict of metadata specified
        - render_context: The Jinja rendering context specified
        - resources: A list of resource definitions

        :param expression: A Python expression that will be evaluated. If it evaluates to a truthy result,
           the component name is included in the returned list.

           Example: ``name == "dummy" and metadata.foo in (1,2,3)``.

           A failing evaluation will be treated as False. An empty expression will always match.
        """
        found = []
        for comp in self.iter_components(filtered=filtered):
            if not expression:
                found.append(comp)
                continue
            try:
                result = eval(expression, {}, comp)
            except Exception:
                result = False
            if result:
                found.append(comp)
        return found

    def filter_components(self, include: str | None = None, exclude: str | None = None):
        """
        Applies a component filter to the project config (report.component_filter.include/exclude).
        Use this to create variants of the same report (think full/light version).
        Include- and Exclude-expressions can be used at the same time for efficient filtering.

        :param include: A regular expression which matches on the component names to be included.
        :param exclude: A regular expression which matches on the component names to be excluded.
        """
        if include is None and exclude is None:
            msg = "No filter expression provided. At least one must be provided."
            raise ValueError(msg)

        filter_include = ".*" if include is None else include
        self.put_setting("report.component_filter.include", filter_include)
        self.put_setting("report.component_filter.exclude", exclude)

    def get_resource(self, alias: str, component: str) -> resource.Resource:
        """
        Finds a Resource from a project component by its alias.

        :param component: The component's name
        :param alias: The resource's alias
        """
        for r in self.iter_resources(component):
            if r["alias"] == alias:
                obj = resource.Resource.from_dict(r)
                obj._cachedir = str(self.sphinx_report_project / ".resource_cache" / component)
                return obj

        msg = f"Cannot find resource {alias!r} in component {component!r}!"
        raise LookupError(msg)

    def iter_resources(self, component: str) -> Iterator[omegaconf.DictConfig]:
        """
        Returns an iterator over all resources from a project component.

        :param component: The component's name
        """
        for comp in self.iter_components():
            if comp["name"] == component:
                yield from comp["resources"]
                return

        msg = f"Component {component} does not exist!"
        raise LookupError(msg)

    def update_resource(self, alias: str, component: str, resource: resource.Resource):
        """
        Updates a Resource from a project component by its alias.

        :param component: The component's name
        :param alias: The resource's alias
        :param resource: The resource's alias
        """
        for comp in self.iter_components():
            if comp["name"] == component:
                for res in comp["resources"]:
                    if res["alias"] == alias:
                        res.update(resource.to_dict())
                        self.save_settings()
                        return
                comp["resources"].append(resource.to_dict())
                self.save_settings()
                return

        msg = f"Component {component} does not exist!"
        raise LookupError(msg)

    def get_default_sphinx_configuration(self, confdir: PathLike):
        """
        Provides Sphinx project configurations, that will be dynamically included by the generated conf.py.
        """
        confdir = Path(confdir)

        config = {
            "pharaoh_jinja_templates": [],
            "pharaoh_jinja_context": {},
            "extensions": [
                "sphinx_design",  # https://github.com/executablebooks/sphinx-design
                "pharaoh.templating.second_level.sphinx_ext.jinja_ext",
                "pharaoh.templating.second_level.sphinx_ext.asset_ext",
            ],
            "project": self.get_setting("report.title"),
            "author": self.get_setting("report.author"),
            "copyright": f"{datetime.datetime.now(tz=datetime.timezone.utc).date()}",
            # The suffix(es) of source filenames.
            # You can specify multiple suffix as a list of string:
            "source_suffix": [".rst", ".rst.jinja2", ".txt"],
            "master_doc": "index",
            "templates_path": ["_templates"],
            # List of patterns, relative to source directory, that match files and
            # directories to ignore when looking for source files.
            # This pattern also affects html_static_path and html_extra_path.
            "exclude_patterns": [".asset_build", "**/asset_scripts"],
            # Make sure the target is unique
            "autosectionlabel_prefix_document": True,
            # Latex Builder (PDF)
        }

        # compute exclusions based on project config
        all_names = {c.name for c in self.iter_components()}
        names_after_filtering = {c.name for c in self.iter_components(filtered=True)}
        excluded_names = all_names - names_after_filtering
        for name in excluded_names:
            config["exclude_patterns"].append(
                (self.sphinx_report_project_components / name).relative_to(self.sphinx_report_project).as_posix()
            )

        # HTML Builder
        static_path = confdir / "_static"
        assert static_path.exists()
        css_files = list((static_path / "css").glob("*.css"))
        js_files = list((static_path / "js").glob("*.js"))
        config.update(
            {
                "html_title": self.get_setting("report.title"),
                "html_use_smartypants": True,
                "html_last_updated_fmt": "%b %d, %Y",
                "html_split_index": False,
                "html_sidebars": {
                    "**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
                },
                # Custom index page
                # https://ofosos.org/2018/12/28/landing-page-template/
                # html_additional_pages={"index": "index.html"},
                # Add any paths that contain custom static files (such as style sheets) here,
                # relative to this directory. They are copied after the builtin static files,
                # so a file named "pharaoh.css" will overwrite the builtin "pharaoh.css".
                "html_static_path": ["_static"],
                "html_css_files": [f"css/{cssfile.name}" for cssfile in css_files],
                "html_js_files": [f"js/{jsfile.name}" for jsfile in js_files],
                "html_show_sphinx": False,
                "html_show_copyright": True,
                "html_context": {
                    "pharaoh_version": pharaoh.__version__,
                    "pharaoh_logo": "_static/html_logo.png",
                    "pharaoh_show_logo": self.get_setting("report.html.show_pharaoh_logo"),
                    "pharaoh_homepage": "",  # todo: change to RTD
                },
                "html_style": "sphinx_rtd_theme_overrides.css",
                "html_theme": "sphinx_rtd_theme",
                "html_theme_options": {
                    "display_version": False,
                    "prev_next_buttons_location": "both",
                    "style_external_links": False,
                    "collapse_navigation": False,
                    "sticky_navigation": True,
                    "navigation_depth": 5,
                    "includehidden": False,
                },
            }
        )

        PM.pharaoh_configure_sphinx(self, config, confdir)

        return config

    def generate_assets(self, component_filters: Iterable[str] = (".*",)) -> list[Path]:
        """
        Generate all assets by executing the asset scripts of a selected or all components.

        All asset scripts are executed in separate parallel child processes
        (number of workers determined by ``asset_gen.worker_processes`` setting;
        setting 0 executes all asset scripts sequentially in the current process).

        Setting ``asset_gen.script_ignore_pattern`` determines if a script is ignored.

        Putting the comment ``# pharaoh: ignore`` at the start of a script will also ignore the file.

        :param component_filters: A list of regular expressions that are matched against each component name.
                                  If a component name matches any of the regular expressions, the component's
                                  assets are regenerated (containing directory will be cleared)
        """
        from pharaoh.assetlib.generation import generate_assets, generate_assets_parallel

        pharaoh.log.log_version_info()
        log.info("Generating assets...")

        PM.pharaoh_asset_gen_started(self)
        sources = []
        for comp in self.iter_components(filtered=True):
            comp_name = comp["name"]
            comp_asset_build_dir = self.asset_build_dir / comp_name
            for cfilter in component_filters:
                if re.match(cfilter, comp_name, re.IGNORECASE) is not None:
                    if comp_asset_build_dir.exists():
                        shutil.rmtree(comp_asset_build_dir, ignore_errors=False)

                    cachedir = self.sphinx_report_project / ".resource_cache" / comp_name
                    log.info(f"Preparing resources of component {comp_name!r} for asset generation...")
                    cachedir.mkdir(parents=True, exist_ok=True)
                    resources: dict[str, resource.Resource] = {
                        resource_dict["alias"]: resource.Resource.from_dict(resource_dict)
                        for resource_dict in comp["resources"]
                    }
                    for r in resources.values():
                        r._cachedir = str(cachedir)

                    PM.pharaoh_asset_gen_prepare_resources(project=self, resources=resources)

                    # TransformedResource may depend on others, so process them first.
                    for r in resources.values():
                        r._cachedir = str(cachedir)
                        if isinstance(r, resource.TransformedResource):
                            r.transform(resources=resources)

                    for script in (self.sphinx_report_project_components / comp_name / "asset_scripts").rglob("*"):
                        sources.append((comp_name, script))
                    break

        workers = self.get_setting("asset_gen.worker_processes", 0)
        if workers == 0:  # Run in same process - used for easier debugging
            results: list[tuple[Path, str | None]] = []
            for component_name, asset_source in sources:
                try:
                    generate_assets(self.project_root, asset_src=asset_source, component_name=component_name)
                    results.append((asset_source, None))
                except Exception:
                    results.append((asset_source, traceback.format_exc()))
        else:
            results = generate_assets_parallel(self.project_root, asset_sources=sources, workers=workers)

        msg = "At least one error occurred while asset script execution:\n"
        i = 1
        processed_asset_scripts = []
        for script, ex in results:
            if ex:
                msg += f"\n\nError #{i}: {ex}\n"
                i += 1
            else:
                processed_asset_scripts.append(script)
        if i > 1:
            pharaoh.log.log_debug_info()
            raise AssetGenerationError(msg)

        return processed_asset_scripts

    def build_report(self, catch_errors=True) -> int:
        """
        Builds the Sphinx project and returns the status code.

        :param catch_errors: If True, Sphinx build errors will not raise an exception but return a -1 instead.
        """
        import multiprocessing

        from sphinx.util.docutils import docutils_namespace, patch_docutils

        from pharaoh.sphinx_app import PharaohSphinx

        builder = self.get_setting("report.builder")
        PM.pharaoh_build_started(self, builder)
        self._check_template_dependencies()

        with chdir(self.sphinx_report_project):
            pharaoh.log.log_version_info()
            log.info(f"Building Sphinx project using {builder.upper()} builder...")

            resolved_settings = omegaconf.OmegaConf.create(
                omegaconf.OmegaConf.merge(
                    self._settings_map["default"],
                    self._settings_map["project"],
                    self._settings_map["env"],
                )
            )
            omegaconf.OmegaConf.resolve(resolved_settings)
            self.sphinx_report_build.mkdir(parents=True, exist_ok=True)
            with open(self.sphinx_report_build / "pharaoh.resolved.yaml", "w") as fp:
                omegaconf.OmegaConf.save(config=resolved_settings, f=fp)

            # Sphinx is used via its API instead of the CLI because we want to modify the info/warning stream
            try:
                sourcedir = "."
                confdir = sourcedir
                outputdir = str(self.sphinx_report_build)
                doctreedir = f"{outputdir}\\.doctrees"
                confoverrides: dict | None = {}
                status = pharaoh.log.SphinxLogRedirector(logging.DEBUG)
                warning = pharaoh.log.SphinxLogRedirector(logging.WARNING)
                freshenv = True
                warningiserror = True
                tags = None
                verbosity = int(self.get_setting("report.verbosity", 0))
                jobs = multiprocessing.cpu_count()
                keep_going = True
                pdb = False

                with patch_docutils(confdir), docutils_namespace():
                    app = PharaohSphinx(
                        sourcedir,
                        confdir,
                        outputdir,
                        doctreedir,
                        builder,
                        confoverrides,
                        status,
                        warning,
                        freshenv,
                        warningiserror,
                        tags,
                        verbosity,
                        jobs,
                        keep_going,
                        pdb,
                    )

                    PM.pharaoh_sphinx_app_inited(app)
                    app.build(force_all=False, filenames=None)
                    if app.statuscode:
                        log.error(
                            f"Sphinx build finished with non-zero exit code {app.statuscode}. "
                            f"See warnings/errors in log above!"
                        )
                    else:
                        log.info(f"Sphinx build finished successfully! Report at:\n{self.sphinx_report_build}")

                    PM.pharaoh_build_finished(self, None)

                    return app.statuscode
            except Exception:
                log.error("Errors occurred during Sphinx build:", exc_info=True)
                pharaoh.log.log_debug_info()

                PM.pharaoh_build_finished(self, sys.exc_info())

                if not catch_errors:
                    raise
                return -1

    def open_report(self):  # pragma: no cover
        """
        Opens the generated report (if possible, e.g. for local HTML reports).
        """
        builder = self.get_setting("report.builder").lower()
        if builder == "html" and os.name == "nt":
            report = self.sphinx_report_build / "index.html"
            if not report.is_file():
                log.warning(f"Could not find HTML report {report}!")
                return
            import webbrowser

            webbrowser.WindowsDefault().open_new_tab(f"file://{report}")
        else:
            log.warning(f"The open_report method is not implemented for builder {builder}!")

    def _archive_report_compressed(self, dest: Path) -> Path:
        """
        Create an archive (zipped) from the build folder.

        :param dest: A destination path to create the archive. Relative paths are relative to the project root.
                     If omitted, the filename will be taken from the ``report.archive_name`` setting.
        :return: The path to the archive
        """
        if not dest.suffix:
            dest = dest / self.get_setting("report.archive_name")

        if dest.exists():
            os.remove(dest)
        dest.parent.mkdir(parents=True, exist_ok=True)

        base_name = dest.parent / dest.stem
        fmt = dest.suffix.replace(".", "")
        shutil.make_archive(str(base_name), fmt, self.sphinx_report_build)
        log.info(f"Created archive at {dest}")

        return dest

    def _archive_report_uncompressed(self, dest: Path) -> Path:
        """
        Create an archive (uncompressed) from the build folder.

        :param dest: A destination path to create the archive. Relative paths are relative to the project root.
                     If omitted, the filename will be taken from the ``report.archive_name`` setting.
        :return: The path to the archive
        """
        # strip file extension: if it exists, it most likely contains the '.zip' suffix which makes no sense
        # in this function
        dest = dest.with_suffix("")

        if dest.exists():
            os.rmdir(dest)

        dest.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree(self.sphinx_report_build, dest)
        log.info(f"Created archive at {dest}")

        return dest

    def archive_report(self, dest: PathLike | None = None, suffix: str | None = None, compression: bool = True) -> Path:
        """
        Create an archive from the build folder.

        :param dest: A destination path to create the archive. Relative paths are relative to the project root.
                     If omitted, the filename will be taken from the ``report.archive_name`` setting.
        :param suffix: Suffix to append to the stem of the archive's filename.
        :param compression: True by default, if set to False, will create an uncompressed folder instead of a zip file.
        :return: The path to the archive
        """
        if not self.sphinx_report_build.exists():
            msg = "Pharaoh report not built yet!"
            raise Exception(msg)
        dest_path = Path(self.get_setting("report.archive_name") if dest is None else dest)

        if not dest_path.is_absolute():
            dest_path = (self.project_root / dest_path).absolute()

        # apply suffix to the stem of the filename, not suffix as in file extension - naming is hard ;)
        if suffix:
            dest_path = dest_path.with_name(f"{dest_path.stem}{suffix}{dest_path.suffix}")

        if compression:
            dest_path = self._archive_report_compressed(dest_path)
        else:
            dest_path = self._archive_report_uncompressed(dest_path)

        return dest_path

    def _check_template_dependencies(self):
        """
        Checks if the used templates of all components are have dependencies to templates that are not used in any
        component.

        E.g. "plugin_abc.template_A" requires an additional component that renders the "plugin_abc.template_B" template,
        in order for all glossary references to work properly.
        """
        l1_templates = PM.pharaoh_collect_l1_templates()
        used_templates = set()
        dependent_templates = set()
        for comp in self.iter_components(filtered=True):
            for template in comp.templates:
                if template in l1_templates:
                    used_templates.add(template)
                    for need in l1_templates[template].needs:
                        dependent_templates.add(need)
        if not dependent_templates.issubset(used_templates):
            msg = (
                f"The used project templates requires the existence of components that render "
                f"following templates: {','.join(dependent_templates - used_templates)}"
            )
            raise ProjectInconsistentError(msg)

    def _ensure_project(
        self,
        templates: list[str],
        template_context: dict[str, Any] | None = None,
        recreate=False,
        custom_settings: str | Path | None = None,
    ):
        """
        Creates a new project by rendering the Sphinx default project at the desired project location.
        """
        self.project_root.mkdir(exist_ok=True, parents=True)
        existing_files = tuple(self.project_root.glob("*"))
        if existing_files:
            if recreate:
                shutil.rmtree(
                    self.project_root, onerror=lambda func, path, exc_info: log.warning(f"Could not delete file {path}")
                )
            else:
                # There are existing files, check if those are the required pharaoh project files/folders
                for path in (
                    self.settings_file,
                    self.project_root,
                    self.sphinx_report_project,
                ):
                    if not path.exists():
                        msg = f"Project structure inconsistent. Missing {path}!"
                        raise ProjectInconsistentError(msg)
                self.asset_build_dir.mkdir(exist_ok=True, parents=True)
                return

        # Create a new project
        self.project_root.mkdir(exist_ok=True, parents=True)
        self.asset_build_dir.mkdir(exist_ok=True, parents=True)

        default_settings = self._load_default_settings()
        if isinstance(custom_settings, (str, Path)):
            custom_settings = omegaconf.OmegaConf.load(custom_settings)  # type: ignore[assignment]
            default_settings = omegaconf.OmegaConf.unsafe_merge(default_settings, custom_settings)  # type: ignore[assignment]

        self.settings_file.write_text(omegaconf.OmegaConf.to_yaml(default_settings, resolve=False))

        gitignore = self.project_root / ".gitignore"
        gitignore_lines = [
            "# Auto-generated by Pharaoh",
            "/log*.txt",
            "/report-build",
            "/report-project/.asset_build",
            "/report-project/.resource_cache",
            "/*.zip",
            "/*.idea",
        ]
        PM.pharaoh_modify_gitignore(gitignore_lines)
        gitignore.write_text("\n".join(gitignore_lines), encoding="utf-8")

        from .templating.first_level.render import render_sphinx_base_project

        render_sphinx_base_project(
            outputdir=self.sphinx_report_project,
            templates=templates,
            context={**(template_context or {}), **{"pharaoh_cli_path": PHARAOH_CLI_PATH}},  # noqa: PIE800
        )

        PM.pharaoh_project_created(self)
        return

    def _load_default_settings(self) -> omegaconf.DictConfig:
        default_settings_collection = PM.pharaoh_collect_default_settings()
        loaded = []
        for default_settings in default_settings_collection:
            if isinstance(default_settings, (str, Path)):
                if not Path(default_settings).exists():
                    msg = f"Custom setting YAMl file {default_settings} does not exist!"
                    raise FileNotFoundError(msg)
                loaded.append(omegaconf.OmegaConf.load(default_settings))
            elif isinstance(default_settings, dict):
                loaded.append(omegaconf.OmegaConf.create(default_settings))
            else:
                msg = (
                    f"Unsupported settings type {type(default_settings)} collected from plugin hook "
                    f"pharaoh_collect_default_settings!"
                )
                raise TypeError(msg)

        return functools.reduce(omegaconf.OmegaConf.unsafe_merge, loaded)  # type: ignore[return-value]

    def _update_template_env(self, env: jinja2.Environment):
        """
        Called by PharaohTemplateEnv.sphinx_config_inited_hook to update the template env by
        project specific Jinja globals or filters

        :param env: the Jinja env to update
        """

        def search_error_assets_global():
            """
            Find all error traceback assets in the project grouped by component name.
            """
            error_assets = {}
            for comp in self.iter_components(filtered=True):
                assets = self.asset_finder.search_assets("asset_type == 'error_traceback'", comp.name)
                if assets:
                    error_assets[comp.name] = assets
            return error_assets

        env.globals["get_setting"] = self.get_setting
        env.globals["search_assets_global"] = self.asset_finder.search_assets
        env.globals["search_error_assets_global"] = search_error_assets_global

    def _build_asset_filepath(self, file: PathLike, component_name: str | None = None) -> Path:
        """
        Returns a new file name inside the asset build directory with the same filename as the input file,
        except its file stem is suffixed with a unique uuid4 hash (first 8 chars).

        E.g. `foo/bar/iris_scatter_plot.html` --> `<asset-build-dir>/<component_name>/iris_scatter_plot_ab8b4081.html`
        """
        file = Path(file)
        try:
            component = component_name or context_stack.get_parent_context("generate_assets")["asset"]["component_name"]
        except Exception:
            component = "unknown_component"
        (self.asset_build_dir / component).mkdir(parents=True, exist_ok=True)
        return self.asset_build_dir / component / f"{file.stem}_{str(uuid.uuid4())[:8]}{file.suffix}"


@attrs.define(frozen=True, slots=False)
class Component:
    name: str = attrs.field(validator=(attr.validators.instance_of(str), attr.validators.matches_re(r"\w+")))
    templates: list[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(str),
            iterable_validator=attrs.validators.instance_of(list),
        )
    )
    render_context: dict = attrs.field(factory=dict)
    resources: list[dict] = attrs.field(factory=list, converter=lambda resources: [r.to_dict() for r in resources])
    metadata: dict = attrs.field(factory=dict)

    def get_render_context(self) -> dict:
        for reserved_key in ("component_name", "resources", "metadata"):
            if reserved_key in self.render_context:
                msg = f"{reserved_key!r} is a reserved context key!"
                raise ValueError(msg)

        return dict(**self.render_context, component_name=self.name, resources=self.resources, metadata=self.metadata)

    @staticmethod
    def from_dictconfig(cfg: omegaconf.DictConfig):
        return Component(**omegaconf.OmegaConf.to_container(cfg, resolve=False))  # type: ignore[arg-type]


if __name__ == "__main__":
    pass



from __future__ import annotations

import ast
import os
import platform
from pathlib import Path

import click

from pharaoh.plugins.plugin_manager import PM

if "PHARAOH.LOGGING.LEVEL" not in os.environ:
    os.environ["PHARAOH.LOGGING.LEVEL"] = "INFO"


def parse_dict(ctx, param, value) -> dict:
    """
    Parses a string with dict notation into a dict object using ast.literal_eval
    """
    if not value:
        return {}
    try:
        obj = ast.literal_eval(value)
    except Exception as e:
        msg = f"Not a valid Python expression {value!r}! Error: {e}!"
        raise click.BadParameter(msg) from None
    if not isinstance(obj, dict):
        msg = "Not a valid dict!"
        raise click.BadParameter(msg)
    return obj


def parse_obj(ctx, param, value):
    """
    Parses a string with into a Python object using ast.literal_eval
    """
    if not value:
        return None
    try:
        obj = ast.literal_eval(value)
    except Exception:
        obj = str(value)
    return obj


def parse_resources(ctx, param, value):
    from pharaoh.assetlib import resource
    from pharaoh.plugins.plugin_manager import PM

    if not value:
        return []
    if not isinstance(value, (list, tuple)):
        value = [value]

    eval_context = PM.pharaoh_collect_resource_types()
    resources = []
    for descriptor in value:
        rsrc = eval(descriptor, {}, eval_context)
        if not isinstance(rsrc, resource.Resource):
            msg = f"{value} did not evaluate to a valid Resource type, but instead to {rsrc!r}!"
            raise TypeError(msg)
        resources.append(rsrc)
    return resources


@click.group(chain=True, name="pharaoh")
@click.option(
    "-p",
    "--path",
    default=".",
    type=click.Path(resolve_path=True, path_type=Path),
    help="The path to the Pharaoh report directory (contains pharaoh.yaml) or the path to "
    "pharaoh.yaml itself. If omitted, the current working directory is used.",
)
@click.pass_context
def cli(ctx, path: Path | None):
    """
    Pharaoh Commandline Interface.

    Usage examples:

    \b
        pharaoh new -t pharaoh_testing.simple -c "{'a': 1}" --settings ../my_settings.yaml
        pharaoh add -n dummy1 -t pharaoh_testing.simple -c "{'test_name':'dummy'}"
        pharaoh generate
        pharaoh build
        pharaoh archive

    Multi-command chaining is also possible (all on one line):

    \b
        pharaoh new add --name dummy1 -t pharaoh_testing.simple -c "{'test_name':'dummy1'}"
         add --name dummy2 -t pharaoh_testing.simple -c "{'test_name':'dummy2'}"
         generate build archive -d "archives/myarchive.zip"
    """
    from pharaoh.project import get_project

    ctx.ensure_object(dict)
    ctx.obj["project_path"] = Path(path or os.getcwd())
    try:
        ctx.obj["project"] = get_project(ctx.obj["project_path"])
    except LookupError:
        ctx.obj["project"] = None


@cli.command()
@click.option("-f", "--force", is_flag=True, default=False, help="Force project overwrite if already exists")
@click.option(
    "-t",
    "--template",
    "templates",
    multiple=True,
    default=("pharaoh.default_project",),
    help="The project templates to use for creating the project files.",
)
@click.option("-c", "--context", help="The Jinja rendering context for the selected template. ", callback=parse_dict)
@click.option(
    "-s",
    "--settings",
    help="A path to a YAML file containing settings to overwrite the default project settings.",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=Path),
)
@click.pass_context
def new(
    ctx,
    force: bool,
    templates: list[str],
    context: dict,
    settings: Path | None,
):
    """
    Creates a new Pharaoh report skeleton.

    Example:

    \b
        pharaoh new -t pharaoh_testing.simple -c "{'a': 1}" --settings ../my_settings.yaml
    """
    from pharaoh.project import PharaohProject

    kwargs = {
        "project_root": Path(ctx.obj["project_path"]),
        "overwrite": force,
        "template_context": dict(context),
        "custom_settings": settings,
    }
    if len(templates):
        kwargs["templates"] = templates
    if settings:
        kwargs["custom_settings"] = settings

    p = PharaohProject(**kwargs)  # type: ignore[arg-type]
    p.save_settings(include_env=True)
    ctx.obj["project"] = p


@cli.command()
@click.option(
    "-n",
    "--name",
    required=True,
    type=str,
    prompt=True,
    help="The name of the component. Must be a valid Python identifier. ",
)
@click.option(
    "-t",
    "--template",
    "templates",
    multiple=True,
    default=("pharaoh.empty",),
    help="The component templates to use for creating the components project files.",
)
@click.option("-c", "--context", help="The Jinja rendering context for the selected templates.", callback=parse_dict)
@click.option(
    "-m",
    "--metadata",
    help="A dictionary of metadata that may be used to find the components.Enter a valid json string",
    callback=parse_dict,
)
@click.option(
    "-r",
    "--resource",
    "resources",
    multiple=True,
    help="A Python expression that evaluates to a valid object of type Resource.",
    callback=parse_resources,
)
@click.option("-i", "--index", default=-1, help="The index to add the component. Default is -1 with means at the end.")
@click.pass_context
def add(
    ctx,
    name: str,
    templates: list[str],
    context: dict,
    resources: list,
    metadata: dict,
    index: int,
):
    """
    Adds a new component to the Pharaoh project.

    Example:

    \b
        pharaoh add -n dummy1 -t pharaoh_testing.simple -c "{'test_name':'dummy'}"
        pharaoh add -n dummy1 -t pharaoh_testing.simple -r "FileResource(alias='foo', pattern='.*')"
    """
    project = ctx.obj["project"]
    project.add_component(
        component_name=name,
        templates=templates,
        render_context=context,
        resources=resources,
        metadata=metadata,
        index=index,
    )


@cli.command()
@click.option(
    "-n",
    "--name",
    required=True,
    type=str,
    prompt=True,
    help="The name of the component. Must be a valid Python identifier. ",
)
@click.option(
    "-a",
    "--alias",
    required=True,
    type=str,
    prompt=True,
    help="The resource alias.",
)
@click.option(
    "-r",
    "--resource",
    "resources",
    help="A Python expression that evaluates to a valid object of type Resource.",
    callback=parse_resources,
)
@click.pass_context
def update_resource(
    ctx,
    name: str,
    alias: str,
    resources,
):
    """
    Updates a Resource from a project component by its alias.

    Examples:

    \b
        pharaoh update-resource -n dummy1 -a foo -r "FileResource(alias='baz', pattern='*')"

    """
    project = ctx.obj["project"]
    project.update_resource(alias=alias, component=name, resource=resources[0])


@cli.command()
@click.option(
    "-n",
    "--name",
    required=True,
    type=str,
    prompt=True,
    help="The name of the component. Must be a valid Python identifier. ",
)
@click.option(
    "-t",
    "--template",
    "templates",
    multiple=True,
    default=("pharaoh.default_project",),
    help="The component templates to use for creating the components project files.",
)
@click.option("-c", "--context", help="The Jinja rendering context for the selected templates.", callback=parse_dict)
@click.pass_context
def add_template(
    ctx,
    name: str,
    templates: list[str],
    context: dict,
):
    """
    Adds additional templates to an existing component, that may overwrite existing files during rendering.

    Example:

    \b
        pharaoh add-template -n dummy1 -t pharaoh_testing.simple -c "{'test_name':'dummy'}"
    """
    project = ctx.obj["project"]
    project.add_template_to_component(component_name=name, templates=templates, render_context=context)


@cli.command()
@click.option(
    "-f",
    "--filter",
    required=True,
    prompt=True,
    type=str,
    help="A case-insensitive component filter. Either a full- or regular expression match, depending on regex option.",
)
@click.option(
    "-r",
    "--regex",
    default=False,
    is_flag=True,
    type=bool,
    help="If True, the filter option will be treated as regular expression. "
    "Components that partially match the regular expression are removed.",
)
@click.pass_context
def remove(
    ctx,
    filter: str,
    regex: bool,
):
    """
    Removes one or multiple existing components.

    If the regular expression (case-insensitive) partially matches the component name.
    To do a full match surround the pattern by ^ and &, e.g. ^my_component&.

    Example:

    \b
        pharaoh remove -f dummy.*
    """
    project = ctx.obj["project"]
    removed = project.remove_component(filter=filter, regex=regex)
    click.echo("Removed components: " + ", ".join(removed))


@cli.command()
@click.argument(
    "key",
    required=True,
    type=str,
)
@click.argument(
    "value",
    required=False,
    default=None,
    type=str,
    callback=parse_obj,
)
@click.pass_context
def env(
    ctx,
    key: str,
    value,
):
    """
    Updates the projects settings.

    Example:

    \b
        pharaoh env foo_A 123               -> foo_a: 123
        pharaoh env foo_B bar               -> foo_b: "bar"
        pharaoh env foo_C "{'baz': 123}"    -> foo_c: {'baz': 123}
        pharaoh env foo_D class             -> foo_d: "class"
        pharaoh env foo_E                   -> foo_e: None

    """
    project = ctx.obj["project"]
    project.put_setting(key, value)
    project.save_settings()


@cli.command()
@click.option(
    "-f",
    "--filter",
    "filters",
    multiple=True,
    type=str,
    help="A list of regular expressions that are matched against each component name. "
    "If a component name matches any of the regular expressions, the component's "
    "assets are regenerated (containing directory will be cleared)",
)
@click.pass_context
def generate(ctx, filters: list[str]):
    """
    Generates assets.
     Either of the entire project or just a selected subset of components.

    Examples:

    \b
        pharaoh generate
        pharaoh generate -f dummy[12]
        pharaoh generate -f dummy1 -f dummy2
    """
    project = ctx.obj["project"]
    if filters:
        project.generate_assets(filters)
    else:
        project.generate_assets()


@cli.command()
@click.pass_context
def build(ctx):
    """
    Executes report generation for the current project.

    Examples:

    \b
        pharaoh build
        pharaoh -p "path/to/my/project" build
    """
    project = ctx.obj["project"]
    status = project.build_report()
    if status != 0:
        msg = f"The Pharaoh build returned with non-zero exit code {status}.\nRefer to the log output for details!"
        raise Exception(msg)


@cli.command()
@click.pass_context
def show_report(ctx):
    """
    Opens the generated report (if possible, e.g. for local HTML reports).

    Examples:

    \b
        pharaoh show-report
        pharaoh -p "path/to/my/project" show-report
    """
    project = ctx.obj["project"]
    project.open_report()


@cli.command()
@click.option(
    "-d",
    "--dest",
    default=None,
    type=click.Path(resolve_path=True, path_type=Path),
    help="Path to the archive. Either a directory or filename with extension zip.",
)
@click.pass_context
def archive(ctx, dest: Path | None):
    """
    Archives the report to a ZIP file.
    """
    project = ctx.obj["project"]
    project.archive_report(dest)


@cli.command()
def info():
    """
    Prints Pharaoh version information, e.g. ``Pharaoh v1.0.0 [Python 3.9.13, Windows-10-10.0.19045-SP0]``
    """
    import pharaoh

    click.echo(f"Pharaoh {pharaoh.__version__} [Python {platform.python_version()}, {platform.platform()}]")


@cli.command()
def print_plugins():
    """
    Prints all available plugins and their installation paths
    """
    from pharaoh.plugins import plugin_manager

    click.echo("Loaded Pharaoh Plugins:")
    for name, module in plugin_manager.PM.pm.list_name_plugin():
        click.echo(f"{name} at {module}")


PM.pharaoh_add_cli_commands(cli)


if __name__ == "__main__":
    cli(prog_name="pharaoh")



from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pharaoh.assetlib.finder import Asset
    from pharaoh.project import PharaohProject
    from pharaoh.sphinx_app import PharaohSphinx


def asset_rel_path_from_build(sphinx_app: PharaohSphinx, template_file: Path, asset: Asset):
    asset.copy_to(sphinx_app.assets_dir)
    return (
        Path(os.path.relpath(sphinx_app.confdir, os.path.dirname(template_file)))
        / sphinx_app.assets_dir.name
        / asset.assetfile.name
    ).as_posix()


def asset_rel_path_from_project(project: PharaohProject, asset: Asset):
    return "/" + asset.assetfile.relative_to(project.asset_build_dir.parent).as_posix()



from __future__ import annotations


def merge(a, b, safe=True, path=None):
    """
    Merges dict b into dict a.

    :param a: Mutable dict. This will be the merge target.
    :param b: This will be the merge source.
    :param safe: If True, existing keys will not be overwritten and a KeyExistsError will be raised.
    :param path: Shall be None for users, used for recursion of the function.
    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], safe, [*path, str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                if safe:
                    msg = f"Conflict at {'.'.join([*path, str(key)])}. Only unique keys are allowed!"
                    raise Exception(msg)
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a



from __future__ import annotations

from pathlib import Path

import attrs

from pharaoh.plugins.contextvar import ContextVar

is_str = attrs.validators.instance_of(str)
is_str_or_path = attrs.validators.instance_of((str, Path))
is_bool = attrs.validators.instance_of(str)


def check_exists(obj, attr, value: str):
    path = Path(value)
    if not path.exists():
        msg = f"Template path {path.as_posix()} does not exist!"
        raise FileNotFoundError(msg)


@attrs.define
class L1Template:
    """
    Declares a level-1/generation-time template.

    :ivar str name: The template's name. Used for referencing: ``<plugin-name>.<template-name>``
    :ivar str|Path path: The path to the template directory
    :ivar list[str] needs: A list of templates this template depends on.
        E.g. ["plugin_abc.template_xyz"]. Pharaoh uses this information to verify if all dependent templates are at least
        used once by any component.
    :ivar list[ContextVar] vars: A list of required context variables and their potential defaults.
    """

    name: str = attrs.field(validator=is_str)
    path: str | Path = attrs.field(converter=Path, validator=check_exists)
    needs: list[str] = attrs.field(validator=attrs.validators.deep_iterable(is_str), factory=list)
    vars: list[ContextVar] = attrs.field(
        validator=attrs.validators.deep_iterable(attrs.validators.instance_of(ContextVar)), factory=list
    )



"""
Backport of contextlib.chdir stdlib class added in Python3.11.
"""

from __future__ import annotations

import os
from contextlib import AbstractContextManager
from pathlib import Path


class chdir(AbstractContextManager):
    """Non thread-safe context manager to change the current working directory."""

    def __init__(self, path):
        self.path = path
        self._old_cwd = None

    def __enter__(self) -> Path:
        self._old_cwd = os.getcwd()
        path = Path(self.path)
        path.mkdir(parents=True, exist_ok=True)
        os.chdir(self.path)
        return path

    def __exit__(self, *excinfo):
        if self._old_cwd:
            os.chdir(self._old_cwd)



from __future__ import annotations

import contextlib
import copy
import os
import re
import textwrap
import traceback
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar

from docutils.parsers.rst import Directive, directives
from docutils.parsers.rst.directives.images import Image
from docutils.statemachine import StateMachine
from sphinx.util import logging

import pharaoh
import pharaoh.project
from pharaoh.templating.second_level.util import asset_rel_path_from_build, asset_rel_path_from_project

from .asset_tmpl import render_asset_template

if TYPE_CHECKING:
    from pharaoh.sphinx_app import PharaohSphinx

logger = logging.getLogger("pharaoh_asset")


class PharaohDirectiveError(Exception):
    pass


def setup(app: PharaohSphinx):
    app.pharaoh_proj = pharaoh.project.PharaohProject(os.path.dirname(app.confdir))
    setup.app = app  # type: ignore[attr-defined]
    setup.config = app.config  # type: ignore[attr-defined]
    setup.confdir = app.confdir  # type: ignore[attr-defined]
    app.add_directive("pharaoh-asset", PharaohAssetDirective)
    app.add_directive("pharao-asset", PharaohAssetDirective)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
        "version": pharaoh.__version__,
    }


def get_app() -> PharaohSphinx:
    return setup.app  # type: ignore[attr-defined]


INFO_TEMPLATE = """
Executing pharaoh-asset directive:
    Source:
        File:    {source_file}:{line_no}
        Section: {section}

    Directive:
{block_text}
"""


ERROR_TEMPLATE = """
Error: {error}

Source:
    File:    {source_file}:{line_no}
    Section: {section}

Directive:
{block_text}

Traceback:
{traceback}
"""


# -------------------------------------
# Options
# -------------------------------------
def _option_boolean(arg):
    if not arg or not arg.strip():
        # no argument given, assume used as a flag
        return True
    if arg.strip().lower() in ("no", "0", "false"):
        return False
    if arg.strip().lower() in ("yes", "1", "true"):
        return True
    msg = f'"{arg}" unknown boolean'
    raise ValueError(msg)


# -------------------------------------
# Directive
# -------------------------------------
class PharaohAssetDirective(Directive):
    """The ``.. pharaoh-asset::`` directive."""

    has_content = True
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec: ClassVar = {
        "optional": _option_boolean,
        "ignore-title": _option_boolean,
        "ignore-description": _option_boolean,
        "index": directives.unchanged,
        "filter": directives.unchanged,
        "image-alt": directives.unchanged,
        "image-height": directives.length_or_unitless,
        "image-width": directives.length_or_percentage_or_unitless,
        "image-scale": directives.nonnegative_int,
        "image-align": Image.align,
        "iframe-width": directives.unchanged,
        "iframe-height": directives.unchanged,
        "components": directives.unchanged,
        "datatable-extended-search": _option_boolean,
        # fixme: for now keep this undocumented until a better solution is found to supply multiple CSS rules for
        # the specific datatable that is rendered (CSS id selector)
        "datatable-style-override": directives.unchanged,
        "template": directives.unchanged,
    }

    def run(self) -> list:
        """Run the plot directive."""
        sphinx_app: PharaohSphinx = get_app()
        rst_file = self.state_machine.document.attributes["source"]
        section = self.state_machine.node.shortrepr()
        # The line number actually refers to the rendered file coming out of 2nd-lvl templating,
        source_file = Path(rst_file + ".rendered").relative_to(sphinx_app.confdir)
        # Find lineno in source file. Here we have to iteratively search the parent items for the first
        # occurrence of a source file/line tuple. Let's hope this implementation detail does not change on
        # Sphinx side, otherwise this code might break :)
        try:
            parent = self.content
            items = parent.items
            while not len(items):
                parent = parent.parent
                items = parent.items
            line_no = items[0][1]
        except Exception:
            line_no = "?"

        msg = INFO_TEMPLATE.format(
            source_file=source_file,
            line_no=line_no,
            section=section,
            block_text=textwrap.indent(self.block_text.strip(), " " * 8),
        )
        logger.verbose(msg)

        try:
            return self._run(source_file, line_no)
        except Exception as e:
            # If there's any error, render a nice error description into the report.
            # It will contain the source file + line number as well as the error message and traceback
            # in a nice collapsible card.
            msg = ERROR_TEMPLATE.format(
                error=repr(e),
                source_file=source_file,
                line_no=line_no,
                section=section,
                block_text=textwrap.indent(self.block_text.strip(), " " * 4),
                traceback=textwrap.indent(traceback.format_exc(limit=3).strip(), " " * 4),
            )
            logger.warning(textwrap.indent(msg, " " * 4))

            assert sphinx_app.pharaoh_te is not None
            error = render_asset_template(
                jinja_env=sphinx_app.pharaoh_te,
                template="error",
                error=repr(e),
                content=msg,
            )
            StateMachine.insert_input(self.state_machine, error.splitlines(keepends=False), source=rst_file)
            # Let the Sphinx build fail (if warnings are errors, which is the default) and show a warning
            self.state_machine.reporter.warning(
                f"{e} (see report for traceback or logs above)", source=source_file, line=line_no
            )
            return []

    def _run(self, source_file: Path, line_no: str) -> list:
        arguments = self.arguments
        content = self.content
        state_machine = self.state_machine
        template_file = state_machine.document.attributes["source"]
        options = self.options
        options = {k.replace("-", "_"): v for k, v in options.items()}

        sphinx_app: PharaohSphinx = setup.app  # type: ignore
        assert sphinx_app.pharaoh_proj is not None
        pharaoh_proj: pharaoh.project.PharaohProject = sphinx_app.pharaoh_proj
        asset_finder = pharaoh_proj.asset_finder

        asset_optional = options.pop("optional", False)
        # Per default, only find assets from the same component as the invoking RST file using the special value _this_
        option_components = options.get("components", "_this_").lower().strip()

        # Find the component name of the currently handled RST file
        try:
            current_component = Path(template_file).relative_to(pharaoh_proj.sphinx_report_project_components).parts[0]
        except Exception:
            current_component = None

        component_selection = set()
        if option_components != "_all_":
            for comp in map(str.strip, option_components.split(",")):
                if comp != "_this_":
                    component_selection.add(comp)
                elif current_component is not None:
                    component_selection.add(current_component)

        # Discover assets by filter
        assets = []
        filter_string = (
            split_filter(" ".join(content))
            + split_filter(options.get("filter", ""))
            + split_filter(" ".join(arguments))
        )
        filter_string = list(set(filter_string))
        if len(filter_string) == 1 and filter_string[0].strip().startswith("__ID__"):
            asset = asset_finder.get_asset_by_id(filter_string[0].strip())
            if asset is not None:
                assets.append(asset)
        else:
            for line in filter_string:
                line = line.strip()
                if line:
                    assets.extend(asset_finder.search_assets(line, list(component_selection)))
        # Filter potential duplicates added by multiple "overlapping" asset filters
        assets = asset_finder.sort_assets(list(set(assets)))
        if not len(assets) and not asset_optional:
            msg = "No assets matched!\n"
            raise Exception(msg)

        # Parse index. Following style is possible: 1,2,4-8,9. Whitespace is ignored.
        index = str(options.pop("index", "")).replace(" ", "")
        indices: list[int] = []
        if index:
            for part in index.split(","):
                if "-" in part:
                    start, stop = part.split("-")
                    indices.extend(range(int(start), int(stop)))
                else:
                    indices.append(int(part))

        lines_to_insert = []
        for iasset, asset in enumerate(assets):
            if indices and iasset not in indices:
                continue
            # Template option priorities:
            #  1. Directive options
            #  2. Asset template options
            #  3. Project defaults

            # 1. Directive options
            template_opts = copy.deepcopy(options)
            # 2. Asset template options
            asset_options = asset.context.get("asset", {})
            for k, v in asset_options.items():
                template_opts.setdefault(k, v)
            #  3. Project defaults
            template_opts.setdefault("iframe_width", pharaoh_proj.get_setting("asset_gen.default_iframe_width"))
            template_opts.setdefault("iframe_height", pharaoh_proj.get_setting("asset_gen.default_iframe_height"))
            template_opts.setdefault(
                "datatable_extended_search", pharaoh_proj.get_setting("asset_gen.default_datatable_extended_search")
            )

            if "template" not in template_opts:
                msg = f"No template selected for asset {asset}!"
                raise PharaohDirectiveError(msg)
            template = template_opts["template"].strip().lower()

            image_opts = {key.replace("image_", ""): val for key, val in options.items() if key.startswith("image_")}

            logger.verbose(f"Rendering {asset} with template {template!r}")

            # Let generated error assets using the "catch_exceptions" function also issue a Sphinx warning
            # to fail report generation.
            if asset.context.get("asset_type", "") == "error_traceback":
                msg = (
                    f"An error occurred during asset generation: {asset.context.get('error_message', '')}. "
                    f"Please search the log for the error message including traceback!"
                )
                if msg not in sphinx_app.seen_warnings:
                    sphinx_app.seen_warnings.add(msg)
                    self.state_machine.reporter.warning(msg)

            assert sphinx_app.pharaoh_te is not None
            content = render_asset_template(
                jinja_env=sphinx_app.pharaoh_te,
                template=template,
                opts=template_opts,
                image_opts=image_opts,
                asset=asset,
                asset_rel_path_from_project=partial(asset_rel_path_from_project, pharaoh_proj),
                asset_rel_path_from_build=partial(asset_rel_path_from_build, sphinx_app, Path(template_file)),
            )

            result = render_asset_template(
                jinja_env=sphinx_app.pharaoh_te,
                template="include_wrapper",
                asset=asset,
                content=content,
                opts=template_opts,
            )

            lines_to_insert.extend(result.split("\n"))
        if lines_to_insert:
            # Sphinx 7.2: #10678: Emit “source-read” events for files read via the include directive.
            # This fix actually removed to need for pre-rendering included files, because a Sphinx source-read event
            # is sent once a file is included via the Include directive and then automatically rendered like all other
            # source files.
            # But in our case it may lead to an infinite recursion. Here's what is think happens:
            # The Include directive monkey patches StateMachine.insert_input to trigger source-read events whenever
            # a file is included.
            # If the included code contains a Pharaoh directive that uses StateMachine.insert_input to add content,
            # it triggers again a source-read event on the same file, leading to an infinity loop.
            # We can come out of this loop by using the original parent implementation ``StateMachine.insert_input``
            # instead of the monkey-patched one.
            StateMachine.insert_input(state_machine, lines_to_insert, source=template_file)

        return []


def split_filter(string: str) -> list[str]:
    # First split by all ; that are not escaped with \
    split_by_semicolon = [substring.strip() for substring in re.split(r"(?<!\\);", string)]

    # Join linebreaks with space
    filter = [" ".join(map(str.strip, part.splitlines())) for part in split_by_semicolon]
    with contextlib.suppress(ValueError):
        filter.remove("")
    return filter



# flake8: noqa: F401,F403
"""
This module contains API functions related to asset generation.
When asset scripts are accessing those API functions, the functions can access the project information by themselves.
"""

# Don't remove unused imports. They may be unused here but maybe in user code!
from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

from pharaoh.assetlib.catch_exceptions import catch_exceptions
from pharaoh.assetlib.context import metadata_context
from pharaoh.assetlib.generation import register_asset, register_templating_context
from pharaoh.assetlib.matlab_engine import Matlab
from pharaoh.assetlib.resource import *

if TYPE_CHECKING:
    from pharaoh.assetlib.finder import AssetFinder


def __get_pharaoh_project():
    import inspect
    from pathlib import Path

    import pharaoh
    from pharaoh.project import get_project

    from .util import is_relative_to

    try:
        return get_project()
    except RuntimeError:
        pass

    pharaoh_src = Path(pharaoh.__file__).parents[1]  # also works for site-packages
    # todo: maybe we have to consider additional paths

    # Find the first frame whose filename is not inside the pharaoh library
    for frame in inspect.stack():
        fname = Path(frame.filename)
        if is_relative_to(fname, pharaoh_src) or "/pharaoh_contrib/" in fname.as_posix():
            continue
        return get_project(Path(frame.filename))
    return None


def get_resource(alias: str, component: str | None = None):
    """
    Get a Resource instance by its component name and resource alias.
    """
    proj = __get_pharaoh_project()
    if not component:
        component = get_current_component()
    return proj.get_resource(alias, component)


def find_components(expression: str = ""):
    """
    Find components by their metadata using an evaluated expression.

    The expression must be a valid Python expression and following local variables may be used:

    - name: The name of the component
    - templates: A list of templates used to render the component
    - metadata: A dict of metadata specified
    - render_context: The Jinja rendering context specified
    - resources: A list of resource definitions

    :param expression: A Python expression that will be evaluated. If it evaluates to a truthy result,
       the component name is included in the returned list.

       Example: ``name == "dummy" and metadata.foo in (1,2,3)``.

       A failing evaluation will be treated as False. An empty expression will always match.
    """
    proj = __get_pharaoh_project()
    return proj.find_components(expression)


def get_current_component() -> str:
    """
    If executed from within a script that is placed inside a Pharaoh component, the function returns the components
    name by analyzing the call stack.
    """
    import inspect

    from pharaoh.assetlib import util

    __get_pharaoh_project()
    # JPY_SESSION_NAME is normally set by Jupyter runtime. In our case we use the nbconvert preprocessor
    # so the variable is not set. Instead the ``pharaoh.assetlib.generation.generate_assets`` function sets
    # this variable.
    jpy_nb = os.getenv("JPY_SESSION_NAME")
    path_iter = Path(jpy_nb).parents if jpy_nb else (frame.filename for frame in inspect.stack())
    return util.get_component_name_via_path_iter(path_iter)


def get_asset_finder() -> AssetFinder:
    """
    Returns the :class:`AssetFinder <pharaoh.assetlib.finder.AssetFinder>` instance for the current project
    """
    proj = __get_pharaoh_project()
    return proj.asset_finder



# flake8: noqa
"""
This module contains Pharaoh project-related API functions.
"""

from pharaoh.assetlib.resource import *
from pharaoh.project import PharaohProject, get_project
from pharaoh.sphinx_app import PharaohSphinx



# mypy: disable-error-code="method-assign"
from __future__ import annotations

import io
import os
from contextlib import contextmanager
from pathlib import Path

try:
    import bokeh.io as bokeh_io
    import bokeh.io.export as bokeh_io_export
    import bokeh.io.saving as bokeh_io_saving
    import bokeh.plotting as bokeh_plotting
    import selenium  # noqa: F401
    from bokeh.core.templates import get_env

    BOKEH_AVAIL = True
except (ImportError, ModuleNotFoundError):
    BOKEH_AVAIL = False

from pharaoh import project
from pharaoh.assetlib.context import context_stack
from pharaoh.assetlib.util import parse_signature

if BOKEH_AVAIL:
    os.environ["BOKEH_LOG_LEVEL"] = "error"

    def init_module():
        """
        This function will ensure that the requirements are fulfilled to export bokeh plots
        as png. bokeh relies on selenium to render interactive content as pngs. selenium in
        turn requires a so called 'webdriver' for any of the major browsers for the rendering.

        It will try to download the MS-Edge webdriver and put it on the PATH of the current
        process.
        https://developer.microsoft.com/en-us/microsoft-edge/tools/webdriver/

        :raises RuntimeError: If msedgedriver.exe is not found in the library root folder.
        """
        from bokeh.io import export, webdriver

        def create_edge_webdriver():
            from selenium.webdriver.edge.options import Options
            from selenium.webdriver.edge.webdriver import WebDriver as Edge

            options = Options()
            if "EXECUTOR_NUMBER" in os.environ:  # Only via Jenkins
                # headless option is only working on Jenkins node. On personal accounts,
                # headless mode for Edge is disabled by IT
                options.add_argument("--headless")
            options.add_argument("--hide-scrollbars")
            options.add_argument("--force-device-scale-factor=1")
            options.add_argument("--force-color-profile=srgb")

            return Edge(options=options)

        def _try_create_chromium_webdriver(*args, **kwargs):
            # Patch for bokeh.io.webdriver._try_create_chromium_webdriver
            # Fallback to creating an Edge webdriver. This should only happen if selenium did not cache
            # the browser + webdriver itself and the user has no internet connection.
            try:
                if "EXECUTOR_NUMBER" in os.environ:
                    return create_edge_webdriver()
                return webdriver.create_chromium_webdriver(*args, **kwargs)
            except Exception:
                return None

        webdriver._try_create_chromium_webdriver = _try_create_chromium_webdriver

        def _log_console(driver) -> None:
            levels = {"WARNING", "ERROR", "SEVERE"}
            try:
                logs = driver.get_log("browser")
            except Exception:
                return
            # loibljoh: Don't know why yet, but the Github runners for ubuntu and Python 3.9 and 3.10 return an integer
            if isinstance(logs, int):
                export.log.warning(f"Browser log was an integer: {logs}")
                return

            messages = [log.get("message") for log in logs if log.get("level") in levels]
            if len(messages) > 0:
                export.log.warning("There were browser warnings and/or errors that may have affected your export")
                for message in messages:
                    export.log.warning(message)

        export._log_console = _log_console

    # memorize the unpatched functions
    vanilla_bokeh_io_show = bokeh_io.show
    vanilla_bokeh_save = bokeh_io_saving.save
    vanilla_bokeh_export_png = bokeh_io_export.export_png
    vanilla_bokeh_export_svg = bokeh_io_export.export_svg
    vanilla_bokeh_io_export_png = bokeh_io.export_png
    vanilla_bokeh_io_export_svg = bokeh_io.export_svg

    @contextmanager
    def patch():
        """
        A context manager that patches all functions of bokeh relevant to display/save figures when executed in a
        pharaoh context. When the program exits the patch-context, the patching will be undone.

        The main goals:
            - retain original function call signature
            - suppress display of figures in GUI or browser
            - alter storage location of figures to the pharaoh build folder
            - deduplicate shared plot resources (js/css files)
            - convert all html plots into static image generation by the flick of a switch
            - if static image output is forced, inject static image parameters via config file
        """
        # switching out the original with the patched functions
        if bokeh_io.show == vanilla_bokeh_io_show:
            bokeh_io.show = patched_bokeh_io_show
            bokeh_plotting.show = patched_bokeh_io_show  # uses identical function as bokeh.io module
            bokeh_io.save = patched_bokeh_save
            bokeh_io_saving.save = patched_bokeh_save
            bokeh_io.export_png = patched_bokeh_export_png
            bokeh_io_export.export_png = patched_bokeh_export_png
            bokeh_io.export_svg = patched_bokeh_export_svg
            bokeh_io_export.export_svg = patched_bokeh_export_svg

        yield

        # undo the patching
        if bokeh_io.show != vanilla_bokeh_io_show:
            bokeh_io.show = vanilla_bokeh_io_show
            bokeh_plotting.show = vanilla_bokeh_io_show  # uses identical function as bokeh.io module
            bokeh_io.save = vanilla_bokeh_save
            bokeh_io_saving.save = vanilla_bokeh_save
            bokeh_io.export_png = vanilla_bokeh_export_png
            bokeh_io_export.export_png = vanilla_bokeh_export_png
            bokeh_io.export_svg = vanilla_bokeh_export_svg
            bokeh_io_export.export_svg = vanilla_bokeh_export_svg

    def patched_bokeh_io_show(*args, **kwargs):
        """
        This is the patched version of ``bokeh.io.showing.show(obj, *args, **kwargs)``.

        Changes:
            - becomes a no-op when executed in a pharaoh context

        docs: https://docs.bokeh.org/en/latest/docs/reference/io.html#bokeh.io.showing.show
        """

    def patched_bokeh_export_png(*args, **kwargs):
        params = parse_signature(obj=vanilla_bokeh_export_png, args=args, kwargs=kwargs)
        file = params["filename"]

        active_app = project.get_project()

        if isinstance(file, io.IOBase) or not Path(file).suffix:
            msg = "The file argument must be a valid file path (including suffix)!"
            raise Exception(msg)

        file_path = active_app._build_asset_filepath(file)
        file_path = file_path.with_suffix(file_path.suffix.lower())
        if file_path.suffix not in (".png", ".jpg", ".jpeg"):
            msg = "Only .png/.jpg/.jpeg file extensions allowed!"
            raise Exception(msg)

        params["filename"] = str(file_path)

        fig: bokeh_plotting.figure = params["obj"]
        assert isinstance(fig, bokeh_plotting.figure)
        with context_stack.new_context(
            context_name="bokeh",
            asset={
                "user_filepath": str(file),
                "file": str(file_path),
                "name": file_path.name,
                "stem": file_path.stem,
                "suffix": file_path.suffix,
                "template": "image",
                "template_opt": {},
            },
        ):
            context_stack.get_parent_context(name="generate_assets")["asset"]["index"] += 1
            vanilla_bokeh_export_png(**params)
            context_stack.dump(file_path)

    def patched_bokeh_export_svg(*args, **kwargs):
        params = parse_signature(obj=vanilla_bokeh_export_svg, args=args, kwargs=kwargs)
        file = params["filename"]

        active_app = project.get_project()

        if isinstance(file, io.IOBase) or not Path(file).suffix:
            msg = "The file argument must be a valid file path (including suffix)!"
            raise Exception(msg)

        file_path = active_app._build_asset_filepath(file)
        file_path = file_path.with_suffix(file_path.suffix.lower())
        if file_path.suffix != ".svg":
            msg = "Only .svg file extension allowed!"
            raise Exception(msg)

        params["filename"] = str(file_path)

        fig: bokeh_plotting.figure = params["obj"]
        assert isinstance(fig, bokeh_plotting.figure)
        with context_stack.new_context(
            context_name="bokeh",
            asset={
                "user_filepath": str(file),
                "file": str(file_path),
                "name": file_path.name,
                "stem": file_path.stem,
                "suffix": file_path.suffix,
                "template": "svg",
                "template_opt": {},
            },
        ):
            context_stack.get_parent_context(name="generate_assets")["asset"]["index"] += 1
            vanilla_bokeh_export_svg(**params)
            context_stack.dump(file_path)

    def patched_bokeh_save(*args, **kwargs):
        """
        This is the patched version of ``bokeh.io.saving.save(*args, **kwargs)`` function.

        Changes:
            - takes kwargs from the pharaoh config file and updates the kwargs passed into the function
            - changes the plots storage destination to be inside pharaoh's asset build folder

        docs: https://docs.bokeh.org/en/latest/docs/reference/io.html#bokeh.io.saving.save
        """
        params = parse_signature(obj=vanilla_bokeh_save, args=args, kwargs=kwargs)
        file = params["filename"]
        params["state"] = None
        params["title"] = ""  # Won't be used by template anyway
        params["resources"] = "cdn"

        active_app = project.get_project()

        if isinstance(file, io.IOBase) or not Path(file).suffix:
            msg = "The file argument must be a valid file path (including suffix)!"
            raise Exception(msg)

        if active_app.get_setting("asset_gen.force_static"):
            coerced_path = Path(file).with_suffix(".png")
            _kwargs = {"obj": params["obj"], "filename": str(coerced_path)}
            patched_bokeh_export_png(**_kwargs)
            return

        file_path = active_app._build_asset_filepath(file)
        file_path = file_path.with_suffix(file_path.suffix.lower())
        if file_path.suffix != ".html":
            msg = "Only .html file extension allowed!"
            raise Exception(msg)

        params["filename"] = str(file_path)
        # Reuse get_env() and provide an already loaded template.
        # If a template is given as string, {% extends base %} will be prepended (base == file.html),
        # which will always add HTML tags to our output, but we JUST want the div to be able to embed later.
        params["template"] = get_env().from_string(
            Path(__file__).with_name("bokeh_embed.html").read_text(encoding="utf-8")
        )

        fig: bokeh_plotting.figure = params["obj"]
        assert isinstance(fig, bokeh_plotting.figure)
        with context_stack.new_context(
            context_name="bokeh",
            asset={
                "user_filepath": str(file),
                "file": str(file_path),
                "name": file_path.name,
                "stem": file_path.stem,
                "suffix": file_path.suffix,
                "template": "raw_html",
                "template_opt": {},
            },
        ):
            context_stack.get_parent_context(name="generate_assets")["asset"]["index"] += 1
            vanilla_bokeh_save(**params)
            context_stack.dump(file_path)

else:

    def init_module():
        pass

    @contextmanager
    def patch():
        yield



from __future__ import annotations

import collections
import importlib
import inspect
from pathlib import Path
from typing import TYPE_CHECKING, Callable, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterable


def parse_signature(obj: str | Callable, args: tuple = (), kwargs: dict | None = None, fromclass: bool = False) -> dict:
    """
    Parses the signature of a function or method and returns a dictionary mapping parameter names to their values.

    This function dynamically imports the specified function or method based on its fully qualified name and then
    parses its signature. It supports both standalone functions and class methods. For class methods, the `fromclass`
    parameter must be set to True. The function or method's parameters are then mapped to their respective values
    based on the provided `args` and `kwargs`, with priority given to `args`.

    :param obj: The fully qualified name of the function or method as a string, or the function/method object itself.
    :param args: A tuple of positional arguments that are passed to the function or method.
    :param kwargs: A dictionary of keyword arguments that are passed to the function or method.
    :param fromclass: A boolean indicating whether the specified function is a method of a class. If True,
      `obj` is expected to be in the format "module.class.function". If False, `obj` is expected to be in the format
      "module.function".
    :returns: A dictionary mapping parameter names to their values as determined by the provided `args` and `kwargs`.
    """
    kwargs = kwargs or {}
    if isinstance(obj, str):
        if fromclass:
            module_name, class_name, function_name = obj.rsplit(".", maxsplit=2)
            mod = importlib.import_module(module_name)
            fun = getattr(getattr(mod, class_name), function_name)
        else:
            module_name, function_name = obj.rsplit(".", maxsplit=1)
            mod = importlib.import_module(module_name)
            fun = getattr(mod, function_name)
    else:
        fun = obj
    signature = inspect.signature(fun)
    mapping = {}
    for i, (pname, parameter) in enumerate(dict(signature.parameters).items()):
        default = None if parameter.default is inspect.Parameter.empty else parameter.default
        try:
            value = args[i]
        except IndexError:
            value = kwargs.get(pname, default)
        mapping[pname] = value
    return mapping


def get_component_name_via_path_iter(path_iterator: Iterable[str]) -> str:
    for path in path_iterator:
        parts = Path(path).parts
        if "asset_scripts" in parts:
            return parts[parts.index("asset_scripts") - 1]

    msg = "Current method was not executed from inside a component (no 'asset_scripts' dir in path)."
    raise LookupError(msg)


def is_relative_to(path: Path, other):
    """Return True if the path is relative to another path or False."""
    # Reimplement here since it was first added in Python 3.9's pathlib
    try:
        path.relative_to(other)
        return True
    except ValueError:
        return False


def dotted_getattr(obj: object, key: str):
    """
    Access any attribute using a dot-notation::

        dotted_getattr(obj, "a.b.c")  # returns obj.a.b.c

    """
    if not isinstance(key, str):
        msg = "Only strings allowed!"
        raise TypeError(msg)
    subkeys = key.split(".")
    for subkey in subkeys:
        obj = getattr(obj, subkey)
    return obj


T = TypeVar("T")


def obj_groupby(
    seq: Iterable[T], key: str, sort_reverse: bool = False, attr: str | None = None, default: str | None = None
) -> dict[str, list[T]]:
    """
    Groups an iterable of objects by a nested attribute.

    :param seq: The iterable of objects to group
    :param key: The nested attribute to use for grouping, e.g. "A.B.C"
    :param sort_reverse: Reverse-sort the keys in the returned dictionary
    :param attr: If omitted, the nested attribute specified by "key" is searched on the items in the input sequence.
                 If given, the nested attribute is searched on a sub-attribute. See examples below::

                    attr: None,  key: "A.B.C", -> obj.A.B.C
                    attr: "A",   key: "B.C",   -> obj.A.B.C
                    attr: "A.B", key: "C",     -> obj.A.B.C

    :param default: If "key" is not an existing attribute, sort the items into this default group.
    :return: A dictionary that maps the group names (values of A.B.C) to a list of items out of the input iterable
    """
    d: dict = collections.defaultdict(lambda: [].append)
    for item in seq:
        accessed_item = item
        if isinstance(attr, str):
            accessed_item = dotted_getattr(item, attr)
        try:
            d[dotted_getattr(accessed_item, key)](item)
        except AttributeError:
            if isinstance(default, str):
                d[default](item)
            else:
                raise

    rv = {}
    for k, v in sorted(d.items(), reverse=sort_reverse):
        rv[k] = v.__self__
    return rv



from __future__ import annotations

import builtins
from typing import Any

import attrs


def str2type(value: str) -> type:
    if isinstance(value, type):
        return value
    if not hasattr(builtins, value):
        msg = f"{value!r} is not a builtin type!"
        raise ValueError(msg)
    item = getattr(builtins, value)
    if not isinstance(item, type):
        msg = f"{value!r} is not a builtin type!"
        raise ValueError(msg)
    return item


@attrs.define(frozen=True)
class ContextVar:
    """
    Declares a context variable for a first-level template.

    :ivar str name: The name of the context variable
    :ivar type type: Its type
    :ivar default: Its default value
    :ivar bool req: If it's required or not
    :ivar str desc: Its description
    """

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: type = attrs.field(validator=attrs.validators.instance_of(type), converter=str2type)
    default: Any = attrs.field()
    req: bool = attrs.field(validator=attrs.validators.instance_of(bool), default=True)
    desc: str = attrs.field(validator=attrs.validators.instance_of(str), default="")



from __future__ import annotations

import re
from pathlib import Path

import omegaconf
from jinja2.exceptions import UndefinedError
from jinja2.utils import is_undefined

DEFAULT = object()


def required(value):
    """
    Raises an UndefinedError if the context variable is undefined.
    The actual name of the Jinja global function is ``req``.

    Example::

        {{ h1(ctx.local.test.test_name|req) }}
    """
    if is_undefined(value):
        raise UndefinedError
    return value


def rep(value) -> str:
    """
    Returns ``repr(value)``. The actual name of the Jinja global function is ``repr``.
    """
    return repr(value)


def or_default(value, default):
    """
    Returns the result of ``value or default``.

    Alias: ``or_d``
    """
    return value or default


def oc_resolve(value: omegaconf.DictConfig):
    """
    Recursively converts an OmegaConf config to a primitive container (dict or list) and returns it.
    """
    conf = value.copy()
    return omegaconf.OmegaConf.to_container(conf, resolve=True)


def oc_get(cfg: omegaconf.DictConfig, key, default=DEFAULT):
    """
    Returns an object inside an omegaconf.DictConfig container.
    If the object does not exist, the default is returned.

    Example::

        {% set plot_title = asset.context|oc_get("plot.title", "") %}

    :param cfg: The omegaconf.DictConfig container to search
    :param key: The name of the object to extract, e.g. ``plot.title``.
    :param default: The default to return if key is not in cfg
    """
    try:
        obj = cfg
        for k in key.split("."):
            ret = getattr(obj, k)
            obj = ret
        return ret
    except Exception:
        if default is DEFAULT:
            msg = f"Not attribute {key} in {cfg}!"
            raise AttributeError(msg) from None
        return default


def exists(path: str) -> bool:
    """
    Returns if path exists.
    """
    return Path(path).exists()


def to_path(path: str) -> Path:
    """
    Returns path as a pathlib.Path instance.
    """
    return Path(path)


def hasattr_(obj, name):
    _hasattr = hasattr(obj, name)
    _haskey = False
    if isinstance(obj, (dict, omegaconf.DictConfig)):
        _haskey = name in obj
    return _hasattr or _haskey


def md2html(text):
    """
    Converts the Markdown text (CommonMark syntax) to HTML.
    """
    import mistletoe

    return mistletoe.markdown(text)


def rsub(s, find: str, replace: str):
    """A non-optimal implementation of a regex filter"""
    return re.sub(find, replace, s)


env_filters = {
    "hasattr": hasattr_,
    "req": required,
    "repr": rep,
    "or_d": or_default,
    "or_default": or_default,
    "oc_resolve": oc_resolve,
    "oc_get": oc_get,
    "bool": bool,
    "exists": exists,
    "to_path": to_path,
    "md2html": md2html,
    "rsub": rsub,
}

# Only document the functions defined in here
to_document = {
    k: func for k, func in env_filters.items() if hasattr(func, "__module__") and func.__module__ == __name__
}



# mypy: disable-error-code="empty-body"
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

import pluggy

from . import NAME

_spec = pluggy.HookspecMarker(NAME)
PathLike = Union[str, Path]

if TYPE_CHECKING:
    from types import TracebackType

    import click

    from pharaoh.api import PharaohProject, Resource
    from pharaoh.sphinx_app import PharaohSphinx

    from .template import L1Template


@_spec
def pharaoh_collect_l1_templates() -> list[L1Template]:
    """
    Collects all first-level templates.

    Example::

        @impl
        def pharaoh_collect_l1_templates():
            return [
                L1Template(
                    name="myorg.default_project",
                    path=Path(__file__).parent / "templates" / "level_1" / "default_project",
                ),
            ]
    """


@_spec
def pharaoh_collect_l2_templates() -> list[PathLike]:
    """
    Collects paths to all second-level template directories.
    """


@_spec
def pharaoh_collect_resource_types() -> list[type[Resource]]:
    """
    Allows registering custom resource types.

    Example::

        @impl
        def pharaoh_collect_resource_types():
            from .resources import RemoteResource

            return [RemoteResource]

    """


@_spec
def pharaoh_asset_gen_started(project: PharaohProject):
    """
    This hook is called when the asset generation is executed.

    :param project: The Pharaoh project instance
    """


@_spec
def pharaoh_asset_gen_prepare_resources(project: PharaohProject, resources: dict[str, Resource]):
    """
    This hook is called before asset generation for each component that is selected.

    Example::

        @impl
        def pharaoh_asset_gen_prepare_resources(project: PharaohProject, resources: dict[str, Resource]):
            from .resources import RemoteResource

            # Transform resources
            for r in resources.values():
                if isinstance(r, RemoteResource):
                    r.download(
                        url=project.get_setting("cloud.url"),
                        auth_credentials=project.get_setting("cloud.auth_credentials", to_container=True) or None,
                    )

    :param project: The Pharaoh project instance
    :param resources: A mapping of all resources that are defined for the current component.
    """


@_spec
def pharaoh_configure_sphinx(project: PharaohProject, config: dict[str, Any], confdir: Path):
    """
    This hook is called at the end of :func:`pharaoh.project.PharaohProject.get_default_sphinx_configuration`
    when Sphinx is reading the ``conf.py`` file.
    It can update the default Sphinx configuration in-place.

    Example::

        @impl
        def pharaoh_configure_sphinx(project: PharaohProject, config: dict, confdir: Path):
            config["extensions"].append("sphinxcontrib.confluencebuilder")
            config["html_style"] = "sphinx_rtd_theme_overrides.css"

    :param project: The Pharaoh project instance
    :param config: The default Sphinx config defined by Pharaoh.
    :param confdir: The parent directory of conf.py
    """


@_spec
def pharaoh_collect_default_settings() -> Path | dict:
    """
    This hook is called when scaffolding a new Pharaoh project.
    It collects all default settings either as a path to a YAML file or a dictionary.
    The collected results are loaded into ``omegaconf.dictconfig.DictConfig`` instances
    and merged in the same order as they are collected (first discovered may get overwritten,
    last discovered always wins).

    Example::

        @impl
        def pharaoh_collect_default_settings():
            return Path(__file__).with_name("my_own_setting.yaml")

    :return: A path to a YAML file or a dictionary
    """


@_spec
def pharaoh_add_cli_commands(cli: click.Group):
    """
    This hook is called after all click CLI commands have been added to add you own CLI commands.
    Example::

        from pharaoh.plugins.api import impl

        @impl
        def pharaoh_add_cli_commands(cli: click.Group):
            @cli.command()
            @click.pass_context
            def my_cli_command(ctx):
                project = ctx.obj["project"]
                project.do_something
    """


@_spec
def pharaoh_build_started(project: PharaohProject, builder: str):
    """
    This hook is called before a Sphinx build is started.

    :param project: The Pharaoh project instance
    :param builder: The name of the configured builder
    """


@_spec
def pharaoh_sphinx_app_inited(app: PharaohSphinx):
    """
    This hook is after the Sphinx application is initialized.

    :param app: The Pharaoh-specific Sphinx application
    """


@_spec
def pharaoh_build_finished(
    project: PharaohProject, exc_info: tuple[type[BaseException], BaseException, TracebackType] | None
):
    """
    This hook is called after a Sphinx build ended with or without error.

    :param project: The Pharaoh project instance
    :param exc_info: An optional value that is returned by sys.exc_info() in case the build raised an Exception.
    """


@_spec
def pharaoh_modify_gitignore(gitignore_lines: list[str]):
    """
    This hook can be used to modify the content on the .gitignore file that is included in the generated Pharaoh
    project inplace.

    :param gitignore_lines: A list of lines to add to the .gitignore file
    """


@_spec
def pharaoh_project_created(project: PharaohProject):
    """
    This hook is called whenever a Pharaoh report has been generated

    :param project: The Pharaoh project instance
    """


@_spec
def pharaoh_find_asset_render_template(template_name: str) -> str | Path:
    """
    This hook is called when an asset template is looked up by name and can customize the search behavior.
    Per default the templates in package ``pharaoh.templating.second_level.sphinx_ext.asset_ext_templates`` are used.

    This template is used to include an asset into a Sphinx document, e.g. the template for a picture asset
    may be called ``image.rst.jinja2`` and would be using the ``image`` directive to embed it.
    If found, the path to the template is returned.

    .. seealso:: :ref:`reference/directive:Asset Templates`

    .. note:: The template's file name should end with ``.jinja2``.

    :param template_name: The name of the template
    """


@_spec
def pharaoh_get_asset_render_template_mappings() -> dict[str, str]:
    """
    This hook maps file extensions of assets to names of templates (see hook ``pharaoh_find_asset_render_template``)
    that are used to embed the asset into the Sphinx document.
    The file extension is only taken into account if no ``template`` metadata key is present in the asset's metadata.

    Pharaoh provides a default mapping (see :ref:`reference/directive:Asset Templates`) which may be updated by this
    hook.
    """
